"""Functions for calculating statistical network metrics."""

import numpy as np

from .. import has_numba
if has_numba:
    from numba import njit

from .metrics_utils import _graph_laplacian


def _network_pearsonr_vectorized(annot1, annot2, weight):
    annot1 = annot1 - np.mean(annot1)
    annot2 = annot2 - np.mean(annot2)
    upper = np.sum(np.multiply(weight, np.outer(annot1, annot2)))
    lower1 = np.sum(np.multiply(weight, np.outer(annot1, annot1)))
    lower2 = np.sum(np.multiply(weight, np.outer(annot2, annot2)))
    return upper / np.sqrt(lower1) / np.sqrt(lower2)


def _network_pearsonr_numba(annot1, annot2, weight):
    n = annot1.shape[0]
    annot1 = annot1 - np.mean(annot1)
    annot2 = annot2 - np.mean(annot2)
    upper, lower1, lower2 = 0.0, 0.0, 0.0
    for i in range(n):
        for j in range(n):
            upper += annot1[i] * annot2[j] * weight[i, j]
            lower1 += annot1[i] * annot1[j] * weight[i, j]
            lower2 += annot2[i] * annot2[j] * weight[i, j]
    return upper / np.sqrt(lower1) / np.sqrt(lower2)


if has_numba:
    _network_pearsonr_numba = njit(_network_pearsonr_numba)


def network_pearsonr(annot1, annot2, weight, use_numba=has_numba):
    r"""
    Calculate pearson correlation between two annotation vectors.

    .. warning::
       Test before use.

    Parameters
    ----------
    annot1 : (N,) array_like
        First annotation vector, demean will be applied.
    annot2 : (N,) array_like
        Second annotation vector, demean will be applied.
    weight : (N, N) array_like
        Weight matrix. Diagonal elements should be 1.
    use_numba : bool, optional
        Whether to use numba for calculation. Default: True
        (if numba is available).

    Returns
    -------
    corr : float
        Network correlation between `annot1` and `annot2`

    Notes
    -----
    If Pearson correlation is represented as

    .. math::
        \rho_{x,y} = \dfrac{
            \mathrm{sum}(I \times (\hat{x} \otimes \hat{y}))
        }{
            \sigma_x \sigma_y
        }

    The network correlation is defined analogously as

    .. math::
        \rho_{x,y,G} = \dfrac{
            \mathrm{sum}(W \times (\hat{x} \otimes \hat{y}))
        }{
            \sigma_{x,W} \sigma_{y,W}
        }

    where :math:`\hat{x}` and :math:`\hat{y}` are the demeaned annotation vectors,

    The weight matrix :math:`W` is used to represent the network structure.
    It is usually in the form of :math:`W = \\exp(-kL)` where :math:`L` is the
    length matrix and :math:`k` is a decay parameter.

    Example using shortest path length as weight

    .. code:: python

        spl, _ = distance_wei_floyd(D) # input should be distance matrix
        spl_wei = 1 / np.exp(spl)
        netcorr = network_pearsonr(annot1, annot2, spl_wei)

    Example using (inverse) effective resistance as weight

    .. code:: python

        R_eff = effective_resistance(W)
        R_eff_norm = R_eff / np.max(R_eff)
        W = 1 / R_eff_norm
        W = W / np.max(W)
        np.fill_diagonal(W, 1.0)
        netcorr = network_pearsonr(annot1, annot2, W)

    References
    ----------
    .. [1] Coscia, M. (2021). Pearson correlations on complex networks.
       Journal of Complex Networks, 9(6), cnab036.
       https://doi.org/10.1093/comnet/cnab036


    See Also
    --------
    netneurotools.stats.network_pearsonr_pairwise
    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _network_pearsonr_numba(annot1, annot2, weight)
    else:
        return _network_pearsonr_vectorized(annot1, annot2, weight)


def _cross_outer(annot_mat):
    """
    Calculate cross outer product of input matrix.

    This functions is only used in `network_pearsonr_pairwise`.

    Parameters
    ----------
    annot_mat : (N, D) array_like
        Input matrix

    Returns
    -------
    cross_outer : (N, N, D, D) numpy.ndarray
        Cross outer product of `annot_mat`
    """
    n_samp, n_feat = annot_mat.shape
    cross_outer = np.empty((n_samp, n_samp, n_feat, n_feat), annot_mat.dtype)
    for a in range(n_samp):
        for b in range(n_samp):
            for c in range(n_feat):
                for d in range(n_feat):
                    cross_outer[a, b, c, d] = annot_mat[a, c] * annot_mat[b, d]
    return cross_outer


if has_numba:
    # ("float64[:,:,:,::1](float64[:,::1])")
    _cross_outer = njit(_cross_outer)


def _multiply_sum(cross_outer, weight):
    """
    Multiply and sum cross outer product.

    This functions is only used in `network_pearsonr_pairwise`.

    Parameters
    ----------
    cross_outer : (N, N, D, D) array_like
        Cross outer product of `annot_mat`
    weight : (D, D) array_like
        Weight matrix

    Returns
    -------
    cross_outer_after : (N, N) numpy.ndarray
        Result of multiplying and summing `cross_outer`
    """
    n_samp, _, n_dim, _ = cross_outer.shape
    cross_outer_after = np.empty((n_samp, n_samp), cross_outer.dtype)
    for i in range(n_samp):
        for j in range(n_samp):
            curr_sum = 0.0
            for k in range(n_dim):
                for l in range(n_dim):  # noqa: E741
                    curr_sum += weight[k, l] * cross_outer[i, j, k, l]
            cross_outer_after[i, j] = curr_sum
    return cross_outer_after


if has_numba:
    # ("float64[:,::1](float64[:,:,:,::1],float64[:,::1])")
    _multiply_sum = njit(_multiply_sum)


def network_pearsonr_pairwise(annot_mat, weight):
    """
    Calculate pairwise network correlation between rows of `annot_mat`.

    .. warning::
       Test before use.

    Parameters
    ----------
    annot_mat : (N, D) array_like
        Input matrix
    weight : (D, D) array_like
        Weight matrix. Diagonal elements should be 1.

    Returns
    -------
    corr_mat : (N, N) numpy.ndarray
        Pairwise network correlation matrix

    Notes
    -----
    This is a faster version of :meth:`netneurotools.stats.network_pearsonr`
    for calculating pairwise network correlation between rows of `annot_mat`.
    Check :meth:`netneurotools.stats.network_pearsonr` for details.

    See Also
    --------
    netneurotools.stats.network_pearsonr
    """
    annot_mat_demean = annot_mat - np.mean(annot_mat, axis=1, keepdims=True)
    if has_numba:
        cross_outer = _cross_outer(annot_mat_demean)
        cross_outer_after = _multiply_sum(cross_outer, weight)
    else:
        # https://stackoverflow.com/questions/24839481/python-matrix-outer-product
        cross_outer = np.einsum('ac,bd->abcd', annot_mat_demean, annot_mat_demean)
        cross_outer_after = np.sum(np.multiply(cross_outer, weight), axis=(2, 3))
    # translating the two lines below in numba does not speed up much
    lower = np.sqrt(np.diagonal(cross_outer_after))
    return cross_outer_after / np.einsum('i,j', lower, lower)


def _onehot_quadratic_form_broadcast(Q_star):
    """
    Calculate one-hot quadratic form of input matrix.

    This functions is only used in `effective_resistance`.

    Parameters
    ----------
    Q_star : (N, N) array_like
        Input matrix

    Returns
    -------
    R_eff : (N, N) numpy.ndarray
        One-hot quadratic form of `Q_star`
    """
    n = Q_star.shape[0]
    R_eff = np.empty((n, n), Q_star.dtype)
    for i in range(n):
        for j in range(n):
            R_eff[i, j] = Q_star[i, i] - Q_star[j, i] - Q_star[i, j] + Q_star[j, j]
    return R_eff


if has_numba:
    # ("float64[:,::1](float64[:,::1])")
    _onehot_quadratic_form_broadcast = njit(_onehot_quadratic_form_broadcast)


def effective_resistance(W, directed=True):
    """
    Calculate effective resistance matrix.

    The effective resistance between two nodes in a graph, often used in the context
    of electrical networks, is a measure that stems from the inverse of the Laplacian
    matrix of the graph.

    .. warning::
       Test before use.

    Parameters
    ----------
    W : (N, N) array_like
        Weight matrix.
    directed : bool, optional
        Whether the graph is directed. This is used to determine whether to turn on
        the :code:`hermitian=True` option in :func:`numpy.linalg.pinv`. When you are
        using a symmetric weight matrix (while real-valued implying hermitian), you
        can set this to False for better performance. Default: True

    Returns
    -------
    R_eff : (N, N) numpy.ndarray
        Effective resistance matrix

    Notes
    -----
    The effective resistance between two nodes :math:`i` and :math:`j` is defined as

    .. math::
        R_{ij} = (e_i - e_j)^T Q^* (e_i - e_j)

    where :math:`Q^*` is the Moore-Penrose pseudoinverse of the Laplacian matrix
    :math:`L` of the graph, and :math:`e_i` is the :math:`i`-th standard basis vector.

    References
    ----------
    .. [1] Ellens, W., Spieksma, F. M., Van Mieghem, P., Jamakovic, A., & Kooij,
       R. E. (2011). Effective graph resistance. Linear Algebra and Its Applications,
       435(10), 2491–2506. https://doi.org/10.1016/j.laa.2011.02.024

    See Also
    --------
    netneurotools.stats.network_polarisation
    """
    L = _graph_laplacian(W)
    Q_star = np.linalg.pinv(L, hermitian=not directed)
    if has_numba:
        R_eff = _onehot_quadratic_form_broadcast(Q_star)
    else:
        Q_star_diag = np.diag(Q_star)
        R_eff = \
            Q_star_diag[:, np.newaxis] \
            - Q_star \
            - Q_star.T \
            + Q_star_diag[np.newaxis, :]
    return R_eff


def _polariz_diff(vec):
    """
    Calculate difference between positive and negative parts of a vector.

    This functions is only used in `network_polarisation`.

    Parameters
    ----------
    vec : (N,) array_like
        Input vector. Must have both positive and negative values.

    Returns
    -------
    vec_diff : (N,) numpy.ndarray
        Difference between positive and negative parts of `vec`
    """
    #
    vec_pos = np.maximum(vec, 0.0)
    vec_pos /= np.max(vec_pos)
    #
    vec_neg = np.minimum(vec, 0.0)
    vec_neg = np.abs(vec_neg)
    vec_neg /= np.max(vec_neg)
    return (vec_pos - vec_neg)


if has_numba:
    _polariz_diff = njit(_polariz_diff)


def _quadratic_form(W, vec_left, vec_right, squared=False):
    """
    Calculate quadratic form :math:`v_{left}^T W v_{right}`.

    Parameters
    ----------
    W : (N, N) array_like
        Input matrix.
    vec_left : (N,) array_like
        Left weight vector.
    vec_right : (N,) array_like
        Right weight vector.
    squared : bool, optional
        Whether to square the input weight matrix. Default: False

    Returns
    -------
    quadratic_form : float
        Quadratic form from `W`, `vec_left`, and `vec_right`
    """
    # [numpy]

    # (vec_left.T @ W @ vec_right)[0, 0]
    # [numba]
    # vec = np.ascontiguousarray(vec[np.newaxis, :])
    n = W.shape[0]
    ret = 0.0
    for i in range(n):
        for j in range(n):
            if squared:
                ret += vec_left[i] * vec_right[j] * W[i, j]**2
            else:
                ret += vec_left[i] * vec_right[j] * W[i, j]
    return ret


if has_numba:
    _quadratic_form = njit(_quadratic_form)


def network_polarisation(vec, W, directed=True):
    r"""
    Calculate polarisation of a vector on a graph.

    Network polarisation is a measure of polizzartion taken into account all the
    three factors below [1]_:

    - how extreme the opinions of the people are
    - how much they organize into echo chambers, and
    - how these echo chambers organize in the network

    .. warning::
       Test before use.

    Parameters
    ----------
    vec : (N,) array_like
        Polarization vector. Must have both positive and negative values. Will be
        normalized between -1 and 1 internally.
    W : (N, N) array_like
        Weight matrix.
    directed : bool, optional
        Whether the graph is directed. This is used to determine whether to turn on
        the :code:`hermitian=True` option in :func:`numpy.linalg.pinv`. When you are
        using a symmetric weight matrix (while real-valued implying hermitian), you
        can set this to False for better performance. Default: True

    Returns
    -------
    polariz : float
        Polarization of `vec` on `W`

    Notes
    -----
    The measure is based on the genralized Eucledian distance, defined as

    .. math::
        \delta_{G, o} = \sqrt{(o^+ - o^-)^T Q^* (o^+ - o^-)}

    where :math:`o^+` and :math:`o^-` are the positive and negative parts of the
    polarization vector, and :math:`Q^*` is the Moore-Penrose pseudoinverse
    of the Laplacian matrix :math:`L` of the graph. Check :func:`effective_resistance`
    for similarity.

    References
    ----------
    .. [1] Hohmann, M., Devriendt, K., & Coscia, M. (2023). Quantifying ideological
       polarization on a network using generalized Euclidean distance. Science Advances,
       9(9), eabq2044. https://doi.org/10.1126/sciadv.abq2044

    See Also
    --------
    netneurotools.stats.effective_resistance
    """
    L = _graph_laplacian(W)
    Q_star = np.linalg.pinv(L, hermitian=not directed)
    diff = _polariz_diff(vec)
    if has_numba:
        polariz_sq = _quadratic_form(Q_star, diff, diff, squared=False)
    else:
        polariz_sq = (diff.T @ Q_star @ diff)
    return np.sqrt(polariz_sq)


def _network_variance_vectorized(vec, D):
    p = vec / np.sum(vec)
    return 0.5 * (p.T @ np.multiply(D, D) @ p)


def _network_variance_numba(vec, D):
    p = vec / np.sum(vec)
    return 0.5 * _quadratic_form(D, p, p, squared=True)


if has_numba:
    _network_variance_numba = njit(_network_variance_numba)


def network_variance(vec, D, use_numba=has_numba):
    r"""
    Calculate variance of a vector on a graph.

    Network variance is a measure of variance taken into account the network
    structure.

    .. warning::
       Test before use.

    Parameters
    ----------
    vec : (N,) array_like
        Input vector. Must be all positive.
        Will be normalized internally as a probability distribution.
    D : (N, N) array_like
        Distance matrix.
    use_numba : bool, optional
        Whether to use numba for calculation. Default: True
        (if numba is available).

    Returns
    -------
    network_variance : float
        Network variance of `vec` on `D`

    Notes
    -----
    The network variance is defined as

    .. math::
        var(p) = \frac{1}{2} \sum_{i, j} p(i) p(j) d^2(i,j)

    where :math:`p` is the probability distribution of `vec`, and :math:`d(i,j)`
    is the distance between node :math:`i` and :math:`j`.

    The distance matrix :math:`D` can make use of effective resistance or its
    square root.

    Example using effective resistance as weight matrix

    .. code:: python

        R_eff = effective_resistance(W)
        netvar = network_variance(vec, R_eff)

    References
    ----------
    .. [1] Devriendt, K., Martin-Gutierrez, S., & Lambiotte, R. (2022).
       Variance and covariance of distributions on graphs. SIAM Review, 64(2),
       343–359. https://doi.org/10.1137/20M1361328

    See Also
    --------
    netneurotools.stats.network_covariance
    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _network_variance_numba(vec, D)
    else:
        return _network_variance_vectorized(vec, D)


def _network_covariance_vectorized(joint_pmat, D, calc_marginal=True):
    p = np.sum(joint_pmat, axis=1)
    q = np.sum(joint_pmat, axis=0)
    D_sq = np.multiply(D, D)
    cov = p.T @ D_sq @ q - np.sum(np.multiply(joint_pmat, D_sq))
    if calc_marginal:
        var_p = p.T @ D_sq @ p
        var_q = q.T @ D_sq @ q
    else:
        var_p, var_q = 0, 0
    return 0.5 * cov, 0.5 * var_p, 0.5 * var_q


def _network_covariance_numba(joint_pmat, D, calc_marginal=True):
    n = joint_pmat.shape[0]
    p = np.sum(joint_pmat, axis=1)
    q = np.sum(joint_pmat, axis=0)
    cov = 0.0
    var_p, var_q = 0.0, 0.0
    for i in range(n):
        for j in range(n):
            cov += (p[i] * q[j] - joint_pmat[i, j]) * D[i, j]**2
            if calc_marginal:
                var_p += p[i] * p[j] * D[i, j]**2
                var_q += q[i] * q[j] * D[i, j]**2
    return 0.5 * cov, 0.5 * var_p, 0.5 * var_q


if has_numba:
    _network_covariance_numba = njit(_network_covariance_numba)


def network_covariance(joint_pmat, D, calc_marginal=True, use_numba=has_numba):
    r"""
    Calculate covariance of a joint probability matrix on a graph.

    .. warning::
       Test before use.

    Parameters
    ----------
    joint_pmat : (N, N) array_like
        Joint probability matrix. Please make sure that it is valid.
    D : (N, N) array_like
        Distance matrix.
    calc_marginal : bool, optional
        Whether to calculate marginal variance. It will be marginally faster if
        :code:`calc_marginal=False` (returning marginal variances as 0). Default: True
    use_numba : bool, optional
        Whether to use numba for calculation. Default: True
        (if numba is available).

    Returns
    -------
    network_covariance : float
        Covariance of `joint_pmat` on `D`
    var_p : float
        Marginal variance of `joint_pmat` on `D`.
        Will be 0 if :code:`calc_marginal=False`
    var_q : float
        Marginal variance of `joint_pmat` on `D`.
        Will be 0 if :code:`calc_marginal=False`

    Notes
    -----
    The network variance is defined as

    .. math::
        cov(P) = \frac{1}{2} \sum_{i, j} [p(i) q(j) - P(i,j)] d^2(i,j)

    where :math:`P` is the joint probability matrix, :math:`p` and :math:`q`
    are the marginal probability distributions of `joint_pmat`, and :math:`d(i,j)`
    is the distance between node :math:`i` and :math:`j`.

    Check :func:`network_variance` for usage.

    References
    ----------
    .. [1] Devriendt, K., Martin-Gutierrez, S., & Lambiotte, R. (2022).
       Variance and covariance of distributions on graphs. SIAM Review, 64(2),
       343–359. https://doi.org/10.1137/20M1361328

    See Also
    --------
    netneurotools.stats.network_variance
    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _network_covariance_numba(joint_pmat, D, calc_marginal)
    else:
        return _network_covariance_vectorized(joint_pmat, D, calc_marginal)



"""Magics on networks."""


from .bct import (
    # routing
    degrees_und, degrees_dir,
    distance_wei_floyd, retrieve_shortest_path,
    navigation_wu, get_navigation_path_length,
    # diffusion
    communicability_bin, communicability_wei,
    path_transitivity, search_information,
    mean_first_passage_time, diffusion_efficiency,
    resource_efficiency_bin, flow_graph,
    # other
    assortativity_und, assortativity_dir,
    matching_ind_und,
    rich_feeder_peripheral
)


from .metrics_utils import (
    _fast_binarize,
    _graph_laplacian,
)


from .spreading import (
    simulate_atrophy
)


from .statistical import (
    network_pearsonr,
    network_pearsonr_pairwise,
    effective_resistance,
    network_polarisation,
    network_variance,
    network_covariance,
)


__all__ = [
    # bct
    'degrees_und', 'degrees_dir',
    'distance_wei_floyd', 'retrieve_shortest_path',
    'navigation_wu', 'get_navigation_path_length',
    'communicability_bin', 'communicability_wei',
    'path_transitivity', 'search_information',
    'mean_first_passage_time', 'diffusion_efficiency',
    'resource_efficiency_bin', 'flow_graph',
    'assortativity_und', 'assortativity_dir',
    'matching_ind_und',
    'rich_feeder_peripheral',
    # metrics_utils
    '_fast_binarize', '_graph_laplacian',
    # spreading
    'simulate_atrophy',
    # statistical
    'network_pearsonr', 'network_pearsonr_pairwise',
    'effective_resistance', 'network_polarisation',
    'network_variance', 'network_covariance',
]



"""Functions for working with GIFTI files."""

import nibabel as nib

from netneurotools.interface.interface_utils import PARCIGNORE


def extract_gifti_labels(gifti_file, parc_ignore=PARCIGNORE):
    """
    Extract vertices and labels from GIFTI file.

    Parameters
    ----------
    gifti_file : str or Path
        Path to GIFTI file.
    parc_ignore : list, optional
        List of labels to ignore.

    Returns
    -------
    surf_data : np.ndarray
        Surface data.
    keys : tuple
        Keys.
    labels : tuple
        Labels.
    """
    surf_data = nib.load(gifti_file).agg_data()
    label_table = nib.load(gifti_file).labeltable
    keys, labels = list(
        zip(
            *[
                label_tuple
                for label_tuple in label_table.get_labels_as_dict().items()
                if label_tuple[1] not in parc_ignore
            ]
        )
    )
    keys = tuple(map(int, keys))
    return surf_data, keys, labels




from . import _version
__version__ = _version.get_versions()['version']

try:
    from numba import njit
    has_numba = True
except ImportError:
    has_numba = False

__all__ = [
    '__version__',
    has_numba
]



"""For testing netneurotools.datasets.datasets_utils functionality."""
import os

import pytest

from netneurotools.datasets import datasets_utils as utils


@pytest.mark.parametrize('dset, expected', [
    ('atl-cammoun2012', ['fsaverage', 'fsaverage5', 'fsaverage6', 'fslr32k',
                         'MNI152NLin2009aSym', 'gcs']),
    ('tpl-conte69', ['url', 'md5']),
    ('tpl-fsaverage', ['fsaverage' + f for f in ['', '3', '4', '5', '6']]),
    ('atl-schaefer2018', ['fsaverage', 'fsaverage6', 'fsaverage6'])
])
def test_get_dataset_info(dset, expected):
    """Test getting dataset info."""
    info = utils._get_dataset_info(dset)
    if isinstance(info, dict):
        assert all(k in info.keys() for k in expected)
    elif isinstance(info, list):
        for f in info:
            assert all(k in f.keys() for k in expected)
    else:
        assert False

    with pytest.raises(KeyError):
        utils._get_dataset_info('notvalid')


def test_get_data_dir(tmp_path):
    """Test getting data directory."""
    data_dir = utils._get_data_dir(tmp_path)
    assert os.path.isdir(data_dir)



"""Functions for calculating permutation test."""

import numpy as np
from sklearn.utils.validation import check_random_state

try:  # scipy >= 1.8.0
    from scipy.stats._stats_py import _chk2_asarray
except ImportError:  # scipy < 1.8.0
    from scipy.stats.stats import _chk2_asarray

from .correlation import efficient_pearsonr


def permtest_1samp(a, popmean, axis=0, n_perm=1000, seed=0):
    """
    Non-parametric equivalent of :py:func:`scipy.stats.ttest_1samp`.

    Generates two-tailed p-value for hypothesis of whether `a` differs from
    `popmean` using permutation tests

    Parameters
    ----------
    a : array_like
        Sample observations
    popmean : float or array_like
        Expected valued in null hypothesis. If array_like then it must have the
        same shape as `a` excluding the `axis` dimension
    axis : int or None, optional
        Axis along which to compute test. If None, compute over the whole array
        of `a`. Default: 0
    n_perm : int, optional
        Number of permutations to assess. Unless `a` is very small along `axis`
        this will approximate a randomization test via Monte Carlo simulations.
        Default: 1000
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Set to None for "randomness".
        Default: 0

    Returns
    -------
    stat : float or numpy.ndarray
        Difference from `popmean`
    pvalue : float or numpy.ndarray
        Non-parametric p-value

    Notes
    -----
    Providing multiple values to `popmean` to run *independent* tests in
    parallel is not currently supported.

    The lowest p-value that can be returned by this function is equal to 1 /
    (`n_perm` + 1).

    Examples
    --------
    >>> from netneurotools import stats
    >>> np.random.seed(7654567)  # set random seed for reproducible results
    >>> rvs = np.random.normal(loc=5, scale=10, size=(50, 2))

    Test if mean of random sample is equal to true mean, and different mean. We
    reject the null hypothesis in the second case and don't reject it in the
    first case.

    >>> stats.permtest_1samp(rvs, 5.0)
    (array([-0.985602  , -0.05204969]), array([0.48551449, 0.95904096]))
    >>> stats.permtest_1samp(rvs, 0.0)
    (array([4.014398  , 4.94795031]), array([0.00699301, 0.000999  ]))

    Example using axis and non-scalar dimension for population mean

    >>> stats.permtest_1samp(rvs, [5.0, 0.0])
    (array([-0.985602  ,  4.94795031]), array([0.48551449, 0.000999  ]))
    >>> stats.permtest_1samp(rvs.T, [5.0, 0.0], axis=1)
    (array([-0.985602  ,  4.94795031]), array([0.51548452, 0.000999  ]))
    """
    a, popmean, axis = _chk2_asarray(a, popmean, axis)
    rs = check_random_state(seed)

    if a.size == 0:
        return np.nan, np.nan

    # ensure popmean will broadcast to `a` correctly
    if popmean.ndim != a.ndim:
        popmean = np.expand_dims(popmean, axis=axis)

    # center `a` around `popmean` and calculate original mean
    zeroed = a - popmean
    true_mean = zeroed.mean(axis=axis) / 1
    abs_mean = np.abs(true_mean)

    # this for loop is not _the fastest_ but is memory efficient
    # the broadcasting alt. would mean storing zeroed.size * n_perm in memory
    permutations = np.ones(true_mean.shape)
    for _ in range(n_perm):
        flipped = zeroed * rs.choice([-1, 1], size=zeroed.shape)  # sign flip
        permutations += np.abs(flipped.mean(axis=axis)) >= abs_mean

    pvals = permutations / (n_perm + 1)  # + 1 in denom accounts for true_mean

    return true_mean, pvals


def permtest_rel(a, b, axis=0, n_perm=1000, seed=0):
    """
    Non-parametric equivalent of :py:func:`scipy.stats.ttest_rel`.

    Generates two-tailed p-value for hypothesis of whether related samples `a`
    and `b` differ using permutation tests

    Parameters
    ----------
    a, b : array_like
        Sample observations. These arrays must have the same shape.
    axis : int or None, optional
        Axis along which to compute test. If None, compute over whole arrays
        of `a` and `b`. Default: 0
    n_perm : int, optional
        Number of permutations to assess. Unless `a` and `b` are very small
        along `axis` this will approximate a randomization test via Monte
        Carlo simulations. Default: 1000
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Set to None for "randomness".
        Default: 0

    Returns
    -------
    stat : float or numpy.ndarray
        Average difference between `a` and `b`
    pvalue : float or numpy.ndarray
        Non-parametric p-value

    Notes
    -----
    The lowest p-value that can be returned by this function is equal to 1 /
    (`n_perm` + 1).

    Examples
    --------
    >>> from netneurotools import stats

    >>> np.random.seed(12345678)  # set random seed for reproducible results
    >>> rvs1 = np.random.normal(loc=5, scale=10, size=500)
    >>> rvs2 = (np.random.normal(loc=5, scale=10, size=500)
    ...         + np.random.normal(scale=0.2, size=500))
    >>> stats.permtest_rel(rvs1, rvs2)  # doctest: +SKIP
    (-0.16506275161572695, 0.8021978021978022)

    >>> rvs3 = (np.random.normal(loc=8, scale=10, size=500)
    ...         + np.random.normal(scale=0.2, size=500))
    >>> stats.permtest_rel(rvs1, rvs3)  # doctest: +SKIP
    (2.40533726097883, 0.000999000999000999)
    """
    a, b, axis = _chk2_asarray(a, b, axis)
    rs = check_random_state(seed)

    if a.shape[axis] != b.shape[axis]:
        raise ValueError('Provided arrays do not have same length along axis')

    if a.size == 0 or b.size == 0:
        return np.nan, np.nan

    # calculate original difference in means
    ab = np.stack([a, b], axis=0)
    if ab.ndim < 3:
        ab = np.expand_dims(ab, axis=-1)
    true_diff = np.squeeze(np.diff(ab, axis=0)).mean(axis=axis) / 1
    abs_true = np.abs(true_diff)

    # idx array
    reidx = list(np.meshgrid(*[range(f) for f in ab.shape], indexing='ij'))

    permutations = np.ones(true_diff.shape)
    for _ in range(n_perm):
        # use this to re-index (i.e., swap along) the first axis of `ab`
        swap = rs.random_sample(ab.shape[:-1]).argsort(axis=axis)
        reidx[0] = np.repeat(swap[..., np.newaxis], ab.shape[-1], axis=-1)
        # recompute difference between `a` and `b` (i.e., first axis of `ab`)
        pdiff = np.squeeze(np.diff(ab[tuple(reidx)], axis=0)).mean(axis=axis)
        permutations += np.abs(pdiff) >= abs_true

    pvals = permutations / (n_perm + 1)  # + 1 in denom accounts for true_diff

    return true_diff, pvals


def permtest_pearsonr(a, b, axis=0, n_perm=1000, resamples=None, seed=0):
    """
    Non-parametric equivalent of :py:func:`scipy.stats.pearsonr`.

    Generates two-tailed p-value for hypothesis of whether samples `a` and `b`
    are correlated using permutation tests

    Parameters
    ----------
    a,b : (N[, M]) array_like
        Sample observations. These arrays must have the same length and either
        an equivalent number of columns or be broadcastable
    axis : int or None, optional
        Axis along which to compute test. If None, compute over whole arrays
        of `a` and `b`. Default: 0
    n_perm : int, optional
        Number of permutations to assess. Unless `a` and `b` are very small
        along `axis` this will approximate a randomization test via Monte
        Carlo simulations. Default: 1000
    resamples : (N, P) array_like, optional
        Resampling array used to shuffle `a` when generating null distribution
        of correlations. This array must have the same length as `a` and `b`
        and should have at least the same number of columns as `n_perm` (if it
        has more then only `n_perm` columns will be used. When not specified a
        standard permutation is used to shuffle `a`. Default: None
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Set to None for "randomness".
        Default: 0

    Returns
    -------
    corr : float or numpyndarray
        Correlations
    pvalue : float or numpy.ndarray
        Non-parametric p-value

    Notes
    -----
    The lowest p-value that can be returned by this function is equal to 1 /
    (`n_perm` + 1).

    Examples
    --------
    >>> from netneurotools import stats

    >>> np.random.seed(12345678)  # set random seed for reproducible results
    >>> x, y = stats.make_correlated_xy(corr=0.1, size=100)
    >>> stats.permtest_pearsonr(x, y)  # doctest: +SKIP
    (0.10032564626876286, 0.3046953046953047)

    >>> x, y = stats.make_correlated_xy(corr=0.5, size=100)
    >>> stats.permtest_pearsonr(x, y)  # doctest: +SKIP
    (0.500040365781984, 0.000999000999000999)

    Also works with multiple columns by either broadcasting the smaller array
    to the larger:

    >>> z = x + np.random.normal(loc=1, size=100)
    >>> stats.permtest_pearsonr(x, np.column_stack([y, z]))
    (array([0.50004037, 0.25843187]), array([0.000999  , 0.01098901]))

    or by using matching columns in the two arrays (e.g., `x` and `y` vs
    `a` and `b`):

    >>> a, b = stats.make_correlated_xy(corr=0.9, size=100)
    >>> stats.permtest_pearsonr(np.column_stack([x, a]), np.column_stack([y, b]))
    (array([0.50004037, 0.89927523]), array([0.000999, 0.000999]))
    """  # noqa
    a, b, axis = _chk2_asarray(a, b, axis)
    rs = check_random_state(seed)

    if len(a) != len(b):
        raise ValueError('Provided arrays do not have same length')

    if a.size == 0 or b.size == 0:
        return np.nan, np.nan

    if resamples is not None:
        if n_perm > resamples.shape[-1]:
            raise ValueError('Number of permutations requested exceeds size '
                             'of resampling array.')

    # divide by one forces coercion to float if ndim = 0
    true_corr = efficient_pearsonr(a, b)[0] / 1
    abs_true = np.abs(true_corr)

    permutations = np.ones(true_corr.shape)
    for perm in range(n_perm):
        # permute `a` and determine whether correlations exceed original
        if resamples is None:
            ap = a[rs.permutation(len(a))]
        else:
            ap = a[resamples[:, perm]]
        permutations += np.abs(efficient_pearsonr(ap, b)[0]) >= abs_true

    pvals = permutations / (n_perm + 1)  # + 1 in denom accounts for true_corr

    return true_corr, pvals



"""For testing netneurotools.stats.permutation_test functionality."""

import pytest
import numpy as np
from netneurotools import stats


@pytest.mark.xfail
def test_permtest_1samp():
    """Test permutation test for one-sample t-test."""
    assert False
    # n1, n2, n3 = 10, 15, 20
    # rs = np.random.RandomState(1234)
    # rvn1 = rs.normal(loc=8, scale=10, size=(n1, n2, n3))

    # t1, p1 = stats.permtest_1samp(rvn1, 1, axis=0)


def test_permtest_rel():
    """Test permutation test for paired samples."""
    dr, pr = -0.0005, 0.4175824175824176
    dpr = ([dr, -dr], [pr, pr])

    rvs1 = np.linspace(1, 100, 100)
    rvs2 = np.linspace(1.01, 99.989, 100)
    rvs1_2D = np.array([rvs1, rvs2])
    rvs2_2D = np.array([rvs2, rvs1])

    # the p-values in these two cases should be consistent
    d, p = stats.permtest_rel(rvs1, rvs2, axis=0, seed=1234)
    assert np.allclose([d, p], (dr, pr))
    d, p = stats.permtest_rel(rvs1_2D.T, rvs2_2D.T, axis=0, seed=1234)
    assert np.allclose([d, p], dpr)

    # but the p-value will differ here because of _how_ we're drawing the
    # random permutations... it would be nice if this was consistent, but as
    # yet i don't have a great idea on how to make that happen without assuming
    # a whole lot about the data
    pr = 0.51248751
    tpr = ([dr, -dr], [pr, pr])
    d, p = stats.permtest_rel(rvs1_2D, rvs2_2D, axis=1, seed=1234)
    assert np.allclose([d, p], tpr)


def test_permtest_pearsonr():
    """Test permutation test for Pearson correlation."""
    np.random.seed(12345678)
    x, y = stats.make_correlated_xy(corr=0.1, size=100)
    r, p = stats.permtest_pearsonr(x, y)
    assert np.allclose([r, p], [0.10032564626876286, 0.3046953046953047])

    x, y = stats.make_correlated_xy(corr=0.5, size=100)
    r, p = stats.permtest_pearsonr(x, y)
    assert np.allclose([r, p], [0.500040365781984, 0.000999000999000999])

    z = x + np.random.normal(loc=1, size=100)
    r, p = stats.permtest_pearsonr(x, np.column_stack([y, z]))
    assert np.allclose(r, np.array([0.50004037, 0.25843187]))
    assert np.allclose(p, np.array([0.000999, 0.01098901]))

    a, b = stats.make_correlated_xy(corr=0.9, size=100)
    r, p = stats.permtest_pearsonr(np.column_stack([x, a]),
                                   np.column_stack([y, b]))
    assert np.allclose(r, np.array([0.50004037, 0.89927523]))
    assert np.allclose(p, np.array([0.000999, 0.000999]))



"""
Functions to simulate atrophy on brain networks with a S.I.R. spreading model.

Python version of SIRsimulator, by Ying-Qiu Zheng:
https://github.com/yingqiuz/SIR_simulator

The original matlab version was originally used in the following paper:

Zheng, Y. Q., Zhang, Y., Yau, Y., Zeighami, Y., Larcher, K.,
       Misic, B., & Dagher, A. (2019). Local vulnerability and global
       connectivity jointly shape neurodegenerative disease propagation.
       PLoS biology, 17(11), e3000495.

This Python version has been used in:

Shafiei, G., Bazinet, V., Dadar, M., Manera, A. L., Collins, D. L.,
        Dagher, A., ... & Ducharme, S. (2023). Network structure and
        transcriptomic vulnerability shape atrophy in frontotemporal dementia.
        Brain, 146(1), 321-336.
"""

import numpy as np
from scipy.stats import norm, zscore


def simulate_atrophy(SC_den, SC_len, seed, roi_sizes, T_total=1000, dt=0.1,
                     p_stay=0.5, v=1, trans_rate=1, init_number=1, GBA=None,
                     SNCA=None, k1=0.5, k=0, FC=None):
    """
    Simulate atrophy on a specified network.

    This is a python version of SIRsimulator, by Ying-Qiu Zheng:
    https://github.com/yingqiuz/SIR_simulator [1]_. This python version was
    first used in [2]_.

    Parameters
    ----------
    SC_den: (n, n) ndarray
        Structural connectivity matrix (strength)
    SC_len: (n, n) ndarray
        Structural connectivity matrix (len)
    seed: int
        ID of the node to be used as a seed for the atrophy process
    roi_sizes: (n,) ndarray:
        Size of each ROIs in the parcellation
    T_total: int
        Total time steps of the function
    dt: float
        Size of each time step
    p_stay: float
        The probability of staying in the same region per unit time
    v: float
        Speed of the atrophy process
    trans_rate: float
        A scalar value controlling the baseline infectivity
    init_number: int
        Number of injected misfolded protein
    GBA: (n,) ndarray
        GBA gene expression (clearance of misfolded protein)
    SNCA: (n,) ndarray
        SNCA gene expression (synthesis of misfolded protein)
    k1: float
        Ratio between weight of atrophy accrual due to accumulation of
        misfolded agends vs. weight of atrophy accrual due to deafferation.
        Must be between 0 and 1
    FC: (n, n) ndarray
        Functional connectivity
    k: float
        weight of functional connectivity

    Returns
    -------
    simulated_atrophy: (n_regions, T_total) ndarray
        Trajectory matrix of the simulated atrophy in individual brain regions.

    References
    ----------
    .. [1] Zheng, Y. Q., Zhang, Y., Yau, Y., Zeighami, Y., Larcher, K.,
       Misic, B., & Dagher, A. (2019). Local vulnerability and global
       connectivity jointly shape neurodegenerative disease propagation.
       PLoS biology, 17(11), e3000495.

    .. [2] Shafiei, G., Bazinet, V., Dadar, M., Manera, A. L.,
       Collins, D. L., Dagher, A., ... & Ducharme, S. (2023). Network structure
       and transcriptomic vulnerability shape atrophy in frontotemporal
       dementia. Brain, 146(1), 321-336.
    """
    # set-up syn_control
    syn_control = roi_sizes

    # Simulated spread of normal proteins
    Pnor0, Rnor0 = _normal_spread(SC_den,
                                  SC_len,
                                  syn_control,
                                  dt=dt,
                                  p_stay=p_stay,
                                  GBA=GBA,
                                  SNCA=SNCA,
                                  k=k,
                                  FC=FC)

    # Simulated spread of misfolded atrophy
    Rnor_all, Rmis_all = _mis_spread(SC_den,
                                     SC_len,
                                     seed,
                                     syn_control,
                                     roi_sizes,
                                     Rnor0.copy(),
                                     Pnor0.copy(),
                                     v=v,
                                     dt=dt,
                                     p_stay=p_stay,
                                     trans_rate=trans_rate,
                                     init_number=init_number,
                                     T_total=T_total,
                                     GBA=GBA,
                                     SNCA=SNCA,
                                     k=k,
                                     FC=FC)

    # Estimate atrophy
    simulated_atrophy = _atrophy(SC_den,
                                 Rnor_all,
                                 Rmis_all,
                                 dt=dt,
                                 k1=k1,
                                 k=k,
                                 FC=FC)

    return simulated_atrophy


def _normal_spread(SC_den, SC_len, syn_control, v=1, dt=0.1, p_stay=0.5,
                   GBA=None, SNCA=None, FC=None, k=0):
    """
    Simulate the spread of normal proteins in a brain network.

    Part 1 of SIRsimulator. SIRsimulator being the original code written by
    Ying-Qiu Zheng in Matlab (https://github.com/yingqiuz/SIR_simulator) for
    her PLoS Biology paper [1]_. This Python version was first used in [2]_.

    Parameters
    ----------
    SC_den: (n, n) ndarray
        Structural connectivity matrix (strength)
    SC_len: (n, n) ndarray
        Structural connectivity matrix (len)
    syn_control: (n,) ndarray
        Parameters specifying in how many voxels proteins can be synthesized
        for each brain regions (region size, i.e., ROIsize)
    v: float
        Speed of the atrophy process. Default: 1
    dt: float
        Size of each time step. Default: 0.1
    p_stay: float
        The probability of staying in the same region per unit time.
        Default: 0.5
    GBA: (n,) ndarray
        GBA gene expression (clearance of misfolded protein). If None, then
        GBA expression is uniformly distributed across brain regions.
        Default: None
    SNCA: (n,) ndarray
        SNCA gene expression (synthesis of misfolded protein)/ If None, then
        SNCA expression is uniformly distributed across brain regions.
        Default: None
    FC: (n, n) ndarray
        Functional connectivity. Default: None
    k: float
        weight of functional connectivity.  Default: 0

    Returns
    -------
    Rnor: (n,) ndarray
         The population of normal agents in regions before pathogenic
         spreading.
    Pnor: (n,) ndarray
        The population of normal agents in edges before pathogenic spreading.

    References
    ----------
    .. [1] Zheng, Y. Q., Zhang, Y., Yau, Y., Zeighami, Y., Larcher, K.,
       Misic, B., & Dagher, A. (2019). Local vulnerability and global
       connectivity jointly shape neurodegenerative disease propagation.
       PLoS biology, 17(11), e3000495.

    .. [2] Shafiei, G., Bazinet, V., Dadar, M., Manera, A. L.,
       Collins, D. L., Dagher, A., ... & Ducharme, S. (2023). Network structure
       and transcriptomic vulnerability shape atrophy in frontotemporal
       dementia. Brain, 146(1), 321-336.
    """
    # Compute basic information
    N_regions = len(SC_len)

    # make sure the diag are zero
    np.fill_diagonal(SC_den, 0)
    np.fill_diagonal(SC_len, 0)

    # Create a Fake FC matrix if FC is none
    if FC is not None:
        np.fill_diagonal(FC, 0)
    else:
        FC = np.zeros((N_regions, N_regions))

    # set probabilities of moving from region i to edge (i,j))
    weights = SC_den * np.exp(k * FC)
    weight_str = weights.sum(axis=0)
    weights = (1 - p_stay) * weights + p_stay * np.diag(weight_str)
    weights = weights / weight_str[:, np.newaxis]

    # convert gene expression scores to probabilities
    if GBA is None or not np.any(GBA):
        clearance_rate = norm.cdf(np.zeros((N_regions)))
    else:
        clearance_rate = norm.cdf(zscore(GBA))

    if SNCA is None or not np.any(SNCA):
        synthesis_rate = norm.cdf(np.zeros((N_regions)))
    else:
        synthesis_rate = norm.cdf(zscore(SNCA))

    # Rnor, Pnor store results of single simulation at each time
    Rnor = np.zeros((N_regions, 1))  # number of normal agent in regions
    Pnor = np.zeros((N_regions, N_regions))  # number of normal agent in paths

    # normal alpha-syn growth
    # fill the network with normal proteins
    iter_max = 1000000000
    for _ in range(iter_max):
        # moving process

        # regions towards paths
        # movDrt stores the number of proteins towards each region. i.e.
        # element in kth row lth col denotes the number of proteins in region k
        # moving towards l
        movDrt = np.repeat(Rnor, N_regions, axis=1) * weights * dt
        np.fill_diagonal(movDrt, 0)

        # paths towards regions
        # update moving
        with np.errstate(divide='ignore', invalid='ignore'):
            movOut = (Pnor * v) / SC_len
            movOut[SC_len == 0] = 0

        Pnor = Pnor - movOut * dt + movDrt
        np.fill_diagonal(Pnor, 0)

        Rtmp = Rnor
        Rnor = Rnor + movOut.sum(axis=0)[:, np.newaxis] * dt - movDrt.sum(axis=1)[:, np.newaxis]  # noqa

        # growth process
        Rnor_cleared = Rnor * (1 - np.exp(-clearance_rate * dt))[:, np.newaxis]
        Rnor_synthesized = ((synthesis_rate * syn_control) * dt)[:, np.newaxis]
        Rnor = Rnor - Rnor_cleared + Rnor_synthesized

        if np.all(abs(Rnor - Rtmp) < 1e-7 * Rtmp):
            break

    return Pnor, Rnor


def _mis_spread(SC_den, SC_len, seed, syn_control, ROIsize, Rnor, Pnor, v=1,
                dt=0.1, p_stay=0.5, trans_rate=1, init_number=1, T_total=1000,
                GBA=None, SNCA=None, return_agents_in_paths=False,
                FC=None, k=0):
    """
    Simulate the spread of misfolded proteins in a brain network.

    Part 2 of SIRsimulator. SIRsimulator being the original code written by
    Ying-Qiu Zheng in Matlab (https://github.com/yingqiuz/SIR_simulator) for
    her PLoS Biology paper [1]_. This Python version was first used in [2]_.

    Parameters
    ----------
    SC_den: (n, n) ndarray
        Structural connectivity matrix (strength)
    SC_len: (n, n) ndarray
        Structural connectivity matrix (len)
    seed: int
        ID of the node to be used as a seed for the atrophy process
    syn_control: (n,) ndarray
        Parameters specifying in how many voxels proteins can be synthesized
        for each brain regions (region size, i.e., ROIsize)
    ROIsize: (n,) ndarray:
        Size of each ROIs in the parcellation
    Rnor: (n,) ndarray
         The population of normal agents in regions before pathogenic
         spreading.
    Pnor: (n,) ndarray
        The population of normal agents in edges before pathogenic spreading.
    v: float
        Speed of the atrophy process
    dt: float
        Size of each time step
    p_stay: float
        The probability of staying in the same region per unit time
    trans_rate: float
        A scalar value controlling the baseline infectivity
    init_number: int
        Number of injected misfolded protein
    T_total: int
        Total time steps of the function
    GBA: (n,) ndarray
        GBA gene expression (clearance of misfolded protein)
    SNCA: (n,) ndarray
        SNCA gene expression (synthesis of misfolded protein)
    return_agents_in_paths: Boolean
        Whether the function should return the number of normal and misfolded
        proteins in each path (edge) of the network. This could be
        memory-consuming. Default: False
    FC: (n, n) ndarray
        Functional connectivity
    k: float
        weight of functional connectivity

    Returns
    -------
    Rnor_all: (n_regions, T_total) ndarray
        Trajectory matrix of the distribution of normal proteins across brain
        regions for each individual time points.
    Rmis_all: (n_regions, T_total) ndarray
        Trajectory matrix of the distribution of misfolded proteins across
        brain regions for each individual time points.
    Pnor_all: (n_regions, n_regions, T_total) ndarray
        Trajectory matrices of the distribution of normal proteins across
        network paths (edges) for each individual time points.
    Pmis_all: (n_regions, n_regions, T_total) ndarray
        Trajectory matrices of the distribution of misfolded proteins across
        network paths (edges) for each individual time points.

    References
    ----------
    .. [1] Zheng, Y. Q., Zhang, Y., Yau, Y., Zeighami, Y., Larcher, K.,
       Misic, B., & Dagher, A. (2019). Local vulnerability and global
       connectivity jointly shape neurodegenerative disease propagation.
       PLoS biology, 17(11), e3000495.

    .. [2] Shafiei, G., Bazinet, V., Dadar, M., Manera, A. L.,
       Collins, D. L., Dagher, A., ... & Ducharme, S. (2023). Network structure
       and transcriptomic vulnerability shape atrophy in frontotemporal
       dementia. Brain, 146(1), 321-336.
    """
    # Compute basic information
    N_regions = len(SC_len)

    # make sure the diag is zero
    np.fill_diagonal(SC_den, 0)
    np.fill_diagonal(SC_len, 0)

    # Create a Fake FC matrix if FC is none
    if FC is not None:
        np.fill_diagonal(FC, 0)
    else:
        FC = np.zeros((N_regions, N_regions))

    # set probabilities of moving from region i to edge (i,j))
    weights = SC_den * np.exp(k * FC)
    weight_str = weights.sum(axis=0)
    weights = (1 - p_stay) * weights + p_stay * np.diag(weight_str)
    weights = weights / weight_str[:, np.newaxis]

    # convert gene expression scores to probabilities
    if GBA is None or not np.any(GBA):
        clearance_rate = norm.cdf(np.zeros((N_regions)))
    else:
        clearance_rate = norm.cdf(zscore(GBA))

    if SNCA is None or not np.any(SNCA):
        synthesis_rate = norm.cdf(np.zeros((N_regions)))
    else:
        synthesis_rate = norm.cdf(zscore(SNCA))

    # store the number of normal/misfoled alpha-syn at each time step
    Rnor_all = np.zeros((N_regions, T_total))
    Rmis_all = np.zeros((N_regions, T_total))

    if return_agents_in_paths:
        Pnor_all = np.zeros((N_regions, N_regions, T_total))
        Pmis_all = np.zeros((N_regions, N_regions, T_total))

    # Rnor, Rmis, Pnor, Pmis store results of single simulation at each time
    Rmis = np.zeros((N_regions, 1))  # nb of misfolded agent in regions
    Pmis = np.zeros((N_regions, N_regions))  # nb of misfolded agent in paths

    # misfolded protein spreading process

    # inject misfolded alpha-syn
    Rmis[seed] = init_number
    for t in range(T_total):
        # moving process

        # normal proteins : region -->> paths
        movDrt_nor = np.repeat(Rnor, N_regions, axis=1) * weights * dt
        np.fill_diagonal(movDrt_nor, 0)

        # normal proteins : path -->> regions
        with np.errstate(invalid='ignore'):
            movOut_nor = (Pnor * v) / SC_len
            movOut_nor[SC_len == 0] = 0

        # misfolded proteins: region -->> paths
        movDrt_mis = np.repeat(Rmis, N_regions, axis=1) * weights * dt
        np.fill_diagonal(movDrt_mis, 0)

        # misfolded proteins: paths -->> regions
        with np.errstate(invalid='ignore'):
            movOut_mis = (Pmis * v) / SC_len
            movOut_mis[SC_len == 0] = 0

        # update regions and paths
        Pnor = Pnor - movOut_nor * dt + movDrt_nor
        np.fill_diagonal(Pnor, 0)
        Rnor = Rnor + movOut_nor.sum(axis=0)[:, np.newaxis] * dt - movDrt_nor.sum(axis=1)[:, np.newaxis]  # noqa

        Pmis = Pmis - movOut_mis * dt + movDrt_mis
        np.fill_diagonal(Pmis, 0)
        Rmis = Rmis + movOut_mis.sum(axis=0)[:, np.newaxis] * dt - movDrt_mis.sum(axis=1)[:, np.newaxis]  # noqa

        Rnor_cleared = Rnor * (1 - np.exp(-clearance_rate * dt))[:, np.newaxis]
        Rnor_synthesized = ((synthesis_rate * syn_control) * dt)[:, np.newaxis]
        Rmis_cleared = Rmis * (1 - np.exp(-clearance_rate * dt))[:, np.newaxis]

        # the probability of getting misfolded
        gamma0 = trans_rate / ROIsize
        misProb = 1 - np.exp(-Rmis * gamma0[:, np.newaxis] * dt)

        # Number of newly infected
        N_misfolded = Rnor * (np.exp(-clearance_rate)[:, np.newaxis]) * misProb

        # Update
        Rnor = Rnor - Rnor_cleared - N_misfolded + Rnor_synthesized
        Rmis = Rmis - Rmis_cleared + N_misfolded

        Rnor_all[:, t] = np.squeeze(Rnor)
        Rmis_all[:, t] = np.squeeze(Rmis)

    if return_agents_in_paths:
        Pnor_all[:, :, t] = Pnor
        Pmis_all[:, :, t] = Pmis

        return Rnor_all, Rmis_all, Pnor_all, Pmis_all

    else:
        return Rnor_all, Rmis_all


def _atrophy(SC_den, Rnor_all, Rmis_all, dt=0.1, k1=0.5, k=0, FC=None):
    """
    Estimate the atrophy from the normal and misfolded proteins distributions.

    This function is inspired by code originally written in Matlab by Ying-Qiu
    Zheng (https://github.com/yingqiuz/SIR_simulator) for her PLoS Biology
    paper [1]_ and was first used in [2]_.

    Parameters
    ----------
    SC_den: (n_regions, n_regions) ndarray
        Structural connectivity matrix (strength).
    Rnor_all: (n_regions, T_total) ndarray
        Trajectory matrix of the distribution of normal protein across brain
        regions for each individual time points.
    Rmis_all: (n_regions, T_total) ndarray
        Trajectory matrix of the distribution of misfolded protein across brain
        regions for each individual time points.
    dt: float
        Size of each time step
    k1: float
        Ratio between weight of atrophy accrual due to accumulation of
        misfolded agends vs. weight of atrophy accrual due to deafferation.
        Must be between 0 and 1
    k: float
        weight of functional connectivity
    FC: (n, n) ndarray
        Functional connectivity

    Returns
    -------
    simulated_atrophy : (n_regions, T_total) ndarray
        Trajectory matrix of the simulated atrophy in individual brain regions.

    References
    ----------
    .. [1] Zheng, Y. Q., Zhang, Y., Yau, Y., Zeighami, Y., Larcher, K.,
       Misic, B., & Dagher, A. (2019). Local vulnerability and global
       connectivity jointly shape neurodegenerative disease propagation.
       PLoS biology, 17(11), e3000495.

    .. [2] Shafiei, G., Bazinet, V., Dadar, M., Manera, A. L.,
       Collins, D. L., Dagher, A., ... & Ducharme, S. (2023). Network structure
       and transcriptomic vulnerability shape atrophy in frontotemporal
       dementia. Brain, 146(1), 321-336.
    """
    # Compute basic information
    N_regions = len(SC_den)

    # Create empty matrix if FC is none
    if FC is not None:
        np.fill_diagonal(FC, 0)
    else:
        FC = np.zeros((N_regions, N_regions))

    ratio = Rmis_all / (Rnor_all + Rmis_all)
    ratio[ratio == np.inf] = 0  # remove possible inf

    # atrophy growth
    k2 = 1 - k1
    weights = SC_den * np.exp(k * FC)
    weights = weights / weights.sum(axis=0)[:, np.newaxis]

    # neuronal loss caused by lack of input from neighbouring regions
    ratio_cum = np.matmul(weights,
                          (1 - np.exp(-ratio * dt)))

    # one time step back
    ratio_cum = np.c_[np.zeros((N_regions, 1)), ratio_cum[:, :-1]]
    ratio_cum = k2 * ratio_cum + k1 * (1 - np.exp(-ratio * dt))

    simulated_atrophy = np.cumsum(ratio_cum, axis=1)

    return simulated_atrophy



"""Functions for making pretty plots and whatnot."""


from .pysurfer_plotters import (
    plot_conte69, plot_fslr, plot_fsaverage, plot_fsvertex
)


from .pyvista_plotters import (
    pv_plot_surface
)


from .mpl_plotters import (
    _grid_communities, _sort_communities,
    plot_point_brain, plot_mod_heatmap,
)


from .color_utils import (
    available_cmaps
)

__all__ = [
    # pysurfer_plotters
    'plot_conte69', 'plot_fslr', 'plot_fsaverage', 'plot_fsvertex',
    # pyvista_plotters
    'pv_plot_surface',
    # mpl_plotters
    '_grid_communities', '_sort_communities',
    'plot_point_brain', 'plot_mod_heatmap',
    # color_utils
    'available_cmaps'
]



"""For testing netneurotools.networks.networks_utils functionality."""

import numpy as np

from netneurotools import networks


def test_get_triu():
    """Test that get_triu returns correct values."""
    arr = np.arange(9).reshape(3, 3)
    assert np.all(networks.get_triu(arr) == np.array([1, 2, 5]))
    assert np.all(networks.get_triu(arr, k=0) == np.array([0, 1, 2, 4, 5, 8]))



"""Transforms for datasets."""

import numpy as np
import nibabel as nib

from netneurotools.interface.cifti import extract_cifti_surface_labels
from netneurotools.interface.freesurfer import extract_annot_labels
from netneurotools.interface.gifti import extract_gifti_labels
from netneurotools.interface.interface_utils import PARCIGNORE


def _check_vertices_to_parcels_parc_file(parc_file, hemi):
    if hemi == "both":
        if not isinstance(parc_file, (tuple, list)):
            if not str(parc_file).endswith("dlabel.nii"):
                raise ValueError(
                    "parc_file should be tuple or list unless it is a dlabel.nii file"
                )
    elif hemi in ["L", "R"]:
        pass
    else:
        raise ValueError(f"Unknown hemisphere: {hemi}")

    def _check_file_type(path):
        if not str(path).endswith(("gii", "dlabel.nii", "annot")):
            raise ValueError(
                "Unsupported parcellation file format, "
                "only gii, dlabel.nii, and annot are supported."
            )

    if isinstance(parc_file, (tuple, list)):
        _check_file_type(parc_file[0])
        _check_file_type(parc_file[1])
    else:
        _check_file_type(parc_file)


def load_surf_parc_file(parc_file, cifti_structure=None, parc_ignore=PARCIGNORE):
    """
    Load surface parcellation file with vertices and labels.

    Parameters
    ----------
    parc_file : str or Path
        Path to parcellation file.
    cifti_structure : str, optional
        CIFTI structure for dlabel.nii files.
    parc_ignore : list, optional
        List of labels to ignore.

    Returns
    -------
    surf_data : np.ndarray
        Surface data.
    keys : tuple
        Keys.
    labels : tuple
        Labels.
    """
    if str(parc_file).endswith("gii"):
        surf_data, keys, labels = extract_gifti_labels(
            parc_file, parc_ignore=parc_ignore
        )
    elif str(parc_file).endswith("dlabel.nii"):
        if cifti_structure is None:
            raise ValueError("cifti_structure must be specified for dlabel.nii files")
        else:
            cifti = nib.load(parc_file)
            surf_data, keys, labels = extract_cifti_surface_labels(
                cifti.get_fdata(),
                cifti.header.get_axis(0),
                cifti.header.get_axis(1),
                cifti_structure,
                parc_ignore=parc_ignore,
            )
    elif str(parc_file).endswith("annot"):
        surf_data, keys, labels = extract_annot_labels(
            parc_file, parc_ignore=parc_ignore
        )
    else:
        raise ValueError(
            "Unsupported parcellation file format, "
            "only gii, dlabel.nii, and annot are supported."
        )
    return surf_data, keys, labels


def _vertices_to_parcels_single_hemi(
    vert_data, parc_file, cifti_structure=None, background=None, parc_ignore=PARCIGNORE
):
    vertices, keys, labels = load_surf_parc_file(
        parc_file, cifti_structure=cifti_structure, parc_ignore=parc_ignore
    )

    if vertices.shape[0] != len(vert_data):
        raise ValueError(
            "Number of vertices in provided annotation files "
            "differs from size of vertex-level data array.\n"
            "    EXPECTED: {} vertices\n"
            "    RECEIVED: {} vertices".format(vertices.shape[0], len(vert_data))
        )

    if background is not None:
        vert_data[vert_data == background] = np.nan

    reduced = []
    for idx in keys:
        found = vertices == idx
        if not found.any():
            reduced.append(np.empty_like(vert_data[0]) * np.nan)
        else:
            reduced.append(np.nanmean(vert_data[found], axis=0))
    return np.array(reduced), keys, labels


def vertices_to_parcels(
    vert_data, parc_file, hemi="both", background=None, parc_ignore=PARCIGNORE
):
    """
    Convert vertex-level data to parcel-level data.

    Parameters
    ----------
    vert_data : np.ndarray or tuple or list
        Vertex-level data.
    parc_file : str or Path or tuple or list
        Path to parcellation file.
    hemi : str, optional
        Hemisphere to process. Can be 'both', 'L', or 'R'.
    background : int, optional
        Background value in vert_data to ignore.
    parc_ignore : list, optional
        List of labels to ignore.

    Returns
    -------
    reduced : np.ndarray or tuple
        Reduced data.
    keys : tuple
        Keys.
    labels : tuple
        Labels.

    Notes
    -----
    If hemi is 'both':

    * ``vert_data`` can be a tuple or list of two arrays. It can also be a single
      array with left then right hemisphere data.
    * ``parc_file`` should be a tuple or list of two paths, unless it is a dlabel.nii
      file, in which case it can be a single path.
    * Returns tuples of reduced data, keys, and labels.

    If hemi is 'L' or 'R':

    * ``vert_data`` should be a single array.
    * ``parc_file`` should be a single path.
    * Returns reduced data, keys, and labels.
    """
    # _check_vertices_to_parcels_parc_file(parc_file, hemi)

    # deal with parc_file
    if hemi == "both":  # both hemispheres
        if isinstance(parc_file, (tuple, list)):
            cifti_structure_lh = cifti_structure_rh = None
        else:
            cifti_structure_lh = "CIFTI_STRUCTURE_CORTEX_LEFT"
            cifti_structure_rh = "CIFTI_STRUCTURE_CORTEX_RIGHT"
            parc_file = (parc_file, parc_file)

        if len(vert_data) != 2:  # not a tuple or list of two arrays
            vert_data = (
                vert_data[: len(vert_data) // 2],
                vert_data[len(vert_data) // 2 :],
            )

        reduced_lh, keys_lh, labels_lh = _vertices_to_parcels_single_hemi(
            vert_data[0],
            parc_file[0],
            cifti_structure=cifti_structure_lh,
            background=background,
            parc_ignore=parc_ignore,
        )
        reduced_rh, keys_rh, labels_rh = _vertices_to_parcels_single_hemi(
            vert_data[1],
            parc_file[1],
            cifti_structure=cifti_structure_rh,
            background=background,
            parc_ignore=parc_ignore,
        )

        reduced = (reduced_lh, reduced_rh)
        keys = (keys_lh, keys_rh)
        labels = (labels_lh, labels_rh)
    else:
        # single hemisphere
        if hemi == "L":
            cifti_structure_lh = "CIFTI_STRUCTURE_CORTEX_LEFT"
        elif hemi == "R":
            cifti_structure_lh = "CIFTI_STRUCTURE_CORTEX_RIGHT"
        else:
            raise ValueError(f"Unknown hemisphere: {hemi}")

        reduced, keys, labels = _vertices_to_parcels_single_hemi(
            vert_data,
            parc_file,
            cifti_structure=cifti_structure_lh,
            background=background,
            parc_ignore=parc_ignore,
        )

    return reduced, keys, labels


def _parcels_to_vertices_single_hemi(
    parc_data, parc_file, cifti_structure=None, fill=np.nan, parc_ignore=PARCIGNORE
):
    vertices, keys, labels = load_surf_parc_file(
        parc_file, cifti_structure=cifti_structure, parc_ignore=parc_ignore
    )

    # check if keys is ordered
    if not np.array_equal(keys, np.sort(keys)):
        raise RuntimeWarning(f"Keys are not ordered: {keys}, {labels}")

    projected = np.full(
        (vertices.shape[0],) + parc_data.shape[1:], fill, dtype=parc_data.dtype
    )
    for i_idx, idx in enumerate(keys):
        found = vertices == idx
        if not found.any():
            continue
        projected[found] = parc_data[i_idx]
    return projected, keys, labels


def parcels_to_vertices(
    parc_data, parc_file, hemi="both", fill=np.nan, parc_ignore=PARCIGNORE
):
    """
    Convert parcel-level data to vertex-level data.

    Parameters
    ----------
    parc_data : np.ndarray or tuple or list
        Parcel-level data.
    parc_file : str or Path or tuple or list
        Path to parcellation file.
    hemi : str, optional
        Hemisphere to process. Can be 'both', 'L', or 'R'.
    fill : int or float, optional
        Fill value for vertices not in parcels. Default is np.nan.
    parc_ignore : list, optional
        List of labels to ignore.

    Returns
    -------
    projected : np.ndarray or tuple
        Projected data.
    keys : tuple
        Keys.
    labels : tuple
        Labels.

    Notes
    -----
    If hemi is 'both':

    * ``parc_data`` can be a tuple or list of two arrays. It can also be a single
      array with left then right hemisphere data.
    * ``parc_file`` should be a tuple or list of two paths, unless it is a dlabel.nii
      file, in which case it can be a single path.
    * Returns tuples of projected data, keys, and labels.

    If hemi is 'L' or 'R':

    * ``parc_data`` should be a single array.
    * ``parc_file`` should be a single path.
    * Returns projected data, keys, and labels.
    """
    if hemi == "both":
        if isinstance(parc_file, (tuple, list)):
            cifti_structure_lh = cifti_structure_rh = None
        else:
            cifti_structure_lh = "CIFTI_STRUCTURE_CORTEX_LEFT"
            cifti_structure_rh = "CIFTI_STRUCTURE_CORTEX_RIGHT"
            parc_file = (parc_file, parc_file)

        projected_lh, keys_lh, labels_lh = _parcels_to_vertices_single_hemi(
            parc_data[0],
            parc_file[0],
            cifti_structure=cifti_structure_lh,
            fill=fill,
            parc_ignore=parc_ignore,
        )
        projected_rh, keys_rh, labels_rh = _parcels_to_vertices_single_hemi(
            parc_data[1],
            parc_file[1],
            cifti_structure=cifti_structure_rh,
            fill=fill,
            parc_ignore=parc_ignore,
        )

        projected = (projected_lh, projected_rh)
        keys = (keys_lh, keys_rh)
        labels = (labels_lh, labels_rh)
    else:
        if hemi == "L":
            cifti_structure_lh = "CIFTI_STRUCTURE_CORTEX_LEFT"
        elif hemi == "R":
            cifti_structure_lh = "CIFTI_STRUCTURE_CORTEX_RIGHT"
        else:
            raise ValueError(f"Unknown hemisphere: {hemi}")

        projected, keys, labels = _parcels_to_vertices_single_hemi(
            parc_data,
            parc_file,
            cifti_structure=cifti_structure_lh,
            fill=fill,
            parc_ignore=parc_ignore,
        )

    return projected, keys, labels



"""Functions for pysurfer-based plotting."""

import os
import numpy as np
import nibabel as nib

from ..datasets import FREESURFER_IGNORE, _get_freesurfer_subjid


def plot_conte69(data, lhlabel, rhlabel, surf='midthickness',
                 vmin=None, vmax=None, colormap='viridis',
                 colorbar=True, num_labels=4, orientation='horizontal',
                 colorbartitle=None, backgroundcolor=(1, 1, 1),
                 foregroundcolor=(0, 0, 0), **kwargs):
    """
    Plot surface `data` on Conte69 Atlas.

    Parameters
    ----------
    data : (N,) array_like
        Surface data for N parcels
    lhlabel : str
        Path to .gii file (generic GIFTI file) containing labels to N/2 parcels
        on the left hemisphere
    rhlabel : str
        Path to .gii file (generic GIFTI file) containing labels to N/2 parcels
        on the right hemisphere
    surf : {'midthickness', 'inflated', 'vinflated'}, optional
        Type of brain surface. Default: 'midthickness'
    vmin : float, optional
        Minimum value to scale the colormap. If None, the min of the data will
        be used. Default: None
    vmax : float, optional
        Maximum value to scale the colormap. If None, the max of the data will
        be used. Default: None
    colormap : str, optional
        Any colormap from matplotlib. Default: 'viridis'
    colorbar : bool, optional
        Wheter to display a colorbar. Default: True
    num_labels : int, optional
        The number of labels to display on the colorbar.
        Available only if colorbar=True. Default: 4
    orientation : str, optional
        Defines the orientation of colorbar. Can be 'horizontal' or 'vertical'.
        Available only if colorbar=True. Default: 'horizontal'
    colorbartitle : str, optional
        The title of colorbar. Available only if colorbar=True. Default: None
    backgroundcolor : tuple of float values with RGB code in [0, 1], optional
        Defines the background color. Default: (1, 1, 1)
    foregroundcolor : tuple of float values with RGB code in [0, 1], optional
        Defines the foreground color (e.g., colorbartitle color).
        Default: (0, 0, 0)
    kwargs : key-value mapping
        Keyword arguments for `mayavi.mlab.triangular_mesh()`

    Returns
    -------
    scene : mayavi.Scene
        Scene object containing plot
    """
    return plot_fslr(data, lhlabel, rhlabel, surf_atlas='conte69',
                     surf_type=surf, vmin=vmin, vmax=vmax, colormap=colormap,
                     colorbar=colorbar, num_labels=num_labels,
                     orientation=orientation, colorbartitle=colorbartitle,
                     backgroundcolor=backgroundcolor,
                     foregroundcolor=foregroundcolor, **kwargs)


def plot_fslr(data, lhlabel, rhlabel, surf_atlas='conte69',
              surf_type='midthickness', vmin=None, vmax=None,
              colormap='viridis', colorbar=True, num_labels=4,
              orientation='horizontal', colorbartitle=None,
              backgroundcolor=(1, 1, 1), foregroundcolor=(0, 0, 0),
              **kwargs):
    """
    Plot surface `data` on a given fsLR32k atlas.

    Parameters
    ----------
    data : (N,) array_like
        Surface data for N parcels
    lhlabel : str
        Path to .gii file (generic GIFTI file) containing labels to N/2 parcels
        on the left hemisphere
    rhlabel : str
        Path to .gii file (generic GIFTI file) containing labels to N/2 parcels
        on the right hemisphere
    surf_atlas: {'conte69', 'yerkes19'}, optional
        Surface atlas on which to plot 'data'. Default: 'conte69'
    surf_type : {'midthickness', 'inflated', 'vinflated'}, optional
        Type of brain surface. Default: 'midthickness'
    vmin : float, optional
        Minimum value to scale the colormap. If None, the min of the data will
        be used. Default: None
    vmax : float, optional
        Maximum value to scale the colormap. If None, the max of the data will
        be used. Default: None
    colormap : str, optional
        Any colormap from matplotlib. Default: 'viridis'
    colorbar : bool, optional
        Wheter to display a colorbar. Default: True
    num_labels : int, optional
        The number of labels to display on the colorbar.
        Available only if colorbar=True. Default: 4
    orientation : str, optional
        Defines the orientation of colorbar. Can be 'horizontal' or 'vertical'.
        Available only if colorbar=True. Default: 'horizontal'
    colorbartitle : str, optional
        The title of colorbar. Available only if colorbar=True. Default: None
    backgroundcolor : tuple of float values with RGB code in [0, 1], optional
        Defines the background color. Default: (1, 1, 1)
    foregroundcolor : tuple of float values with RGB code in [0, 1], optional
        Defines the foreground color (e.g., colorbartitle color).
        Default: (0, 0, 0)
    kwargs : key-value mapping
        Keyword arguments for `mayavi.mlab.triangular_mesh()`

    Returns
    -------
    scene : mayavi.Scene
        Scene object containing plot
    """
    from ..datasets import fetch_conte69, fetch_yerkes19
    try:
        from mayavi import mlab
    except ImportError:
        raise ImportError('Cannot use plot_fslr() if mayavi is not '
                          'installed. Please install mayavi and try again.') from None

    opts = dict()
    opts.update(**kwargs)

    try:
        if surf_atlas == 'conte69':
            surface = fetch_conte69()[surf_type]
        elif surf_atlas == 'yerkes19':
            surface = fetch_yerkes19()[surf_type]

    except KeyError:
        raise ValueError('Provided surf "{}" is not valid. Must be one of '
                         '[\'midthickness\', \'inflated\', \'vinflated\']'
                         .format(surf_type)) from None

    lhsurface, rhsurface = [nib.load(s) for s in surface]

    lhlabels = nib.load(lhlabel).darrays[0].data
    rhlabels = nib.load(rhlabel).darrays[0].data
    lhvert, lhface = [d.data for d in lhsurface.darrays]
    rhvert, rhface = [d.data for d in rhsurface.darrays]

    # add NaNs for medial wall
    data = np.append(np.nan, data)

    # get lh and rh data
    lhdata = np.squeeze(data[lhlabels.astype(int)])
    rhdata = np.squeeze(data[rhlabels.astype(int)])

    # plot
    lhplot = mlab.figure()
    rhplot = mlab.figure()
    lhmesh = mlab.triangular_mesh(lhvert[:, 0], lhvert[:, 1], lhvert[:, 2],
                                  lhface, figure=lhplot, colormap=colormap,
                                  mask=np.isnan(lhdata), scalars=lhdata,
                                  vmin=vmin, vmax=vmax, **opts)
    lhmesh.module_manager.scalar_lut_manager.lut.nan_color = [0.863, 0.863,
                                                              0.863, 1]
    lhmesh.update_pipeline()
    if colorbar is True:
        mlab.colorbar(title=colorbartitle, nb_labels=num_labels,
                      orientation=orientation)
    rhmesh = mlab.triangular_mesh(rhvert[:, 0], rhvert[:, 1], rhvert[:, 2],
                                  rhface, figure=rhplot, colormap=colormap,
                                  mask=np.isnan(rhdata), scalars=rhdata,
                                  vmin=vmin, vmax=vmax, **opts)
    rhmesh.module_manager.scalar_lut_manager.lut.nan_color = [0.863, 0.863,
                                                              0.863, 1]
    rhmesh.update_pipeline()
    if colorbar is True:
        mlab.colorbar(title=colorbartitle, nb_labels=num_labels,
                      orientation=orientation)
    mlab.view(azimuth=180, elevation=90, distance=450, figure=lhplot)
    mlab.view(azimuth=180, elevation=-90, distance=450, figure=rhplot)

    mlab.figure(bgcolor=backgroundcolor, fgcolor=foregroundcolor,
                figure=lhplot)
    mlab.figure(bgcolor=backgroundcolor, fgcolor=foregroundcolor,
                figure=rhplot)

    return lhplot, rhplot


def plot_fsaverage(data, *, lhannot, rhannot, order='lr', mask=None,
                   noplot=None, subject_id='fsaverage', subjects_dir=None,
                   vmin=None, vmax=None, **kwargs):
    """
    Plot `data` to fsaverage brain using `annot` as parcellation.

    Parameters
    ----------
    data : (N,) array_like
        Data for `N` parcels as defined in `annot`
    lhannot : str
        Filepath to .annot file containing labels to parcels on the left
        hemisphere. If a full path is not provided the file is assumed to
        exist inside the `subjects_dir`/`subject`/label directory.
    rhannot : str
        Filepath to .annot file containing labels to parcels on the right
        hemisphere. If a full path is not provided the file is assumed to
        exist inside the `subjects_dir`/`subject`/label directory.
    order : str, optional
        Order of the hemispheres in the data vector (either 'LR' or 'RL').
        Default: 'LR'
    mask : (N,) array_like, optional
        Binary array where entries indicate whether values in `data` should be
        masked from plotting (True = mask; False = show). Default: None
    noplot : list, optional
        List of names in `lhannot` and `rhannot` to not plot. It is assumed
        these are NOT present in `data`. By default 'unknown' and
        'corpuscallosum' will never be plotted if they are present in the
        provided annotation files. Default: None
    subject_id : str, optional
        Subject ID to use; must be present in `subjects_dir`. Default:
        'fsaverage'
    subjects_dir : str, optional
        Path to FreeSurfer subject directory. If not set, will inherit from
        the environmental variable $SUBJECTS_DIR. Default: None
    vmin : float, optional
        Minimum value for colorbar. If not provided, a robust estimation will
        be used from values in `data`. Default: None
    vmax : float, optional
        Maximum value for colorbar. If not provided, a robust estimation will
        be used from values in `data`. Default: None
    kwargs : key-value pairs
        Provided directly to :func:`~.plot_fsvertex` without modification.

    Returns
    -------
    brain : surfer.Brain
        Plotted PySurfer brain

    Examples
    --------
    >>> import numpy as np
    >>> from netneurotools.datasets import fetch_cammoun2012, \
                                           fetch_schaefer2018
    >>> from netneurotools.plotting import plot_fsaverage

    Plotting with the Cammoun 2012 parcellation we specify `order='RL'` because
    many of the Lausanne connectomes have data for the right hemisphere before
    the left hemisphere.

    >>> values = np.random.rand(219)
    >>> scale = 'scale125'
    >>> cammoun = fetch_cammoun2012('fsaverage', verbose=False)[scale]
    >>> plot_fsaverage(values, order='RL',
    ...                lhannot=cammoun.lh, rhannot=cammoun.rh) # doctest: +SKIP

    Plotting with the Schaefer 2018 parcellation we can use the default
    parameter for `order`:

    >>> values = np.random.rand(400)
    >>> scale = '400Parcels7Networks'
    >>> schaefer = fetch_schaefer2018('fsaverage', verbose=False)[scale]
    >>> plot_fsaverage(values,
    ...                lhannot=schaefer.lh,
    ...                rhannot=schaefer.rh)  # doctest: +SKIP

    """
    def _decode_list(vals):
        """List decoder."""
        return [val.decode() if hasattr(val, 'decode') else val for val in vals]

    subject_id, subjects_dir = _get_freesurfer_subjid(subject_id, subjects_dir)

    # cast data to float (required for NaNs)
    data = np.asarray(data, dtype='float')

    order = order.lower()
    if order not in ('lr', 'rl'):
        raise ValueError('order must be either \'lr\' or \'rl\'')

    if mask is not None and len(mask) != len(data):
        raise ValueError('Provided mask must be the same length as data.')

    if vmin is None:
        vmin = np.nanpercentile(data, 2.5)
    if vmax is None:
        vmax = np.nanpercentile(data, 97.5)

    # parcels that should not be included in parcellation
    drop = FREESURFER_IGNORE.copy()
    if noplot is not None:
        if isinstance(noplot, str):
            noplot = [noplot]
        drop += list(noplot)
    drop = _decode_list(drop)

    vtx_data = []
    for annot, hemi in zip((lhannot, rhannot), ('lh', 'rh')):
        # loads annotation data for hemisphere, including vertex `labels`!
        if not annot.startswith(os.path.abspath(os.sep)):
            annot = os.path.join(subjects_dir, subject_id, 'label', annot)
        labels, _, names = nib.freesurfer.read_annot(annot)
        names = _decode_list(names)

        # get appropriate data, accounting for hemispheric asymmetry
        currdrop = np.intersect1d(drop, names)
        if hemi == 'lh':
            if order == 'lr':
                split_id = len(names) - len(currdrop)
                ldata, rdata = np.split(data, [split_id])
                if mask is not None:
                    lmask, rmask = np.split(mask, [split_id])
            elif order == 'rl':
                split_id = len(data) - len(names) + len(currdrop)
                rdata, ldata = np.split(data, [split_id])
                if mask is not None:
                    rmask, lmask = np.split(mask, [split_id])
        hemidata = ldata if hemi == 'lh' else rdata

        # our `data` don't include the "ignored" parcels but our `labels` do,
        # so we need to account for that. find the label ids that correspond to
        # those and set them to NaN in the `data vector`
        inds = sorted([names.index(n) for n in currdrop])
        for i in inds:
            hemidata = np.insert(hemidata, i, np.nan)
        vtx = hemidata[labels]

        # let's also mask data, if necessary
        if mask is not None:
            maskdata = lmask if hemi == 'lh' else rmask
            maskdata = np.insert(maskdata, inds - np.arange(len(inds)), np.nan)
            vtx[maskdata[labels] > 0] = np.nan

        vtx_data.append(vtx)

    brain = plot_fsvertex(np.hstack(vtx_data), order='lr', mask=None,
                          subject_id=subject_id, subjects_dir=subjects_dir,
                          vmin=vmin, vmax=vmax, **kwargs)

    return brain


def plot_fsvertex(data, *, order='lr', surf='pial', views='lat',
                  vmin=None, vmax=None, center=None, mask=None,
                  colormap='viridis', colorbar=True, alpha=0.8,
                  label_fmt='%.2f', num_labels=3, size_per_view=500,
                  subject_id='fsaverage', subjects_dir=None, data_kws=None,
                  **kwargs):
    """
    Plot vertex-wise `data` to fsaverage brain.

    Parameters
    ----------
    data : (N,) array_like
        Data for `N` parcels as defined in `annot`
    order : {'lr', 'rl'}, optional
        Order of the hemispheres in the data vector. Default: 'lr'
    surf : str, optional
        Surface on which to plot data. Default: 'pial'
    views : str or list, optional
        Which views to plot of brain. Default: 'lat'
    vmin : float, optional
        Minimum value for colorbar. If not provided, a robust estimation will
        be used from values in `data`. Default: None
    vmax : float, optional
        Maximum value for colorbar. If not provided, a robust estimation will
        be used from values in `data`. Default: None
    center : float, optional
        Center of colormap, if desired. Default: None
    mask : (N,) array_like, optional
        Binary array where entries indicate whether values in `data` should be
        masked from plotting (True = mask; False = show). Default: None
    colormap : str, optional
        Which colormap to use for plotting `data`. Default: 'viridis'
    colorbar : bool, optional
        Whether to display the colorbar in the plot. Default: True
    alpha : [0, 1] float, optional
        Transparency of plotted `data`. Default: 0.8
    label_fmt : str, optional
        Format of colorbar labels. Default: '%.2f'
    number_of_labels : int, optional
        Number of labels to display on colorbar. Default: 3
    size_per_view : int, optional
        Size, in pixels, of each frame in the plotted display. Default: 1000
    subjects_dir : str, optional
        Path to FreeSurfer subject directory. If not set, will inherit from
        the environmental variable $SUBJECTS_DIR. Default: None
    subject : str, optional
        Subject ID to use; must be present in `subjects_dir`. Default:
        'fsaverage'
    data_kws : dict, optional
        Keyword arguments for Brain.add_data()

    Returns
    -------
    brain : surfer.Brain
        Plotted PySurfer brain
    """
    # hold off on imports until
    try:
        from surfer import Brain
    except ImportError:
        raise ImportError('Cannot use plot_fsaverage() if pysurfer is not '
                          'installed. Please install pysurfer and try again.') from None

    subject_id, subjects_dir = _get_freesurfer_subjid(subject_id, subjects_dir)

    # cast data to float (required for NaNs)
    data = np.asarray(data, dtype='float')

    # handle data_kws if None
    if data_kws is None:
        data_kws = {}

    if mask is not None and len(mask) != len(data):
        raise ValueError('Provided mask must be the same length as data.')

    order = order.lower()
    if order not in ['lr', 'rl']:
        raise ValueError('Specified order must be either \'lr\' or \'rl\'')

    if vmin is None:
        vmin = np.nanpercentile(data, 2.5)
    if vmax is None:
        vmax = np.nanpercentile(data, 97.5)

    # set up brain views
    if not isinstance(views, (np.ndarray, list)):
        views = [views]

    # size of window will depend on # of views provided
    size = (size_per_view * 2, size_per_view * len(views))
    brain_kws = dict(background='white', size=size)
    brain_kws.update(**kwargs)
    brain = Brain(subject_id=subject_id, hemi='split', surf=surf,
                  subjects_dir=subjects_dir, views=views, **brain_kws)

    hemis = ('lh', 'rh') if order == 'lr' else ('rh', 'lh')
    for n, (hemi, vtx_data) in enumerate(zip(hemis, np.split(data, 2))):
        # let's mask data, if necessary
        if mask is not None:
            maskdata = np.asarray(np.split(mask, 2)[n], dtype=bool)
            vtx_data[maskdata] = np.nan

        # we don't want NaN values plotted so set a threshold if they exist
        thresh, nanmask = None, np.isnan(vtx_data)
        if np.any(nanmask) > 0:
            thresh = np.nanmin(vtx_data) - 1
            vtx_data[nanmask] = thresh
            thresh += 0.5

        # finally, add data to this hemisphere!
        brain.add_data(vtx_data, vmin, vmax, hemi=hemi, mid=center,
                       thresh=thresh, alpha=1.0, remove_existing=False,
                       colormap=colormap, colorbar=colorbar, verbose=False,
                       **data_kws)

        if alpha != 1.0:
            surf = brain.data_dict[hemi]['surfaces']
            for _, s in enumerate(surf):
                s.actor.property.opacity = alpha
                s.render()

        # if we have a colorbar, update parameters accordingly
        if colorbar:
            # update label format, as desired
            surf = brain.data_dict[hemi]['surfaces']
            cmap = brain.data_dict[hemi]['colorbars']
            # this updates the format of the colorbar labels
            if label_fmt is not None:
                for n, cm in enumerate(cmap):
                    cm.scalar_bar.label_format = label_fmt
                    surf[n].render()
            # this updates how many labels are shown on the colorbar
            if num_labels is not None:
                for n, cm in enumerate(cmap):
                    cm.scalar_bar.number_of_labels = num_labels
                    surf[n].render()

    return brain



"""Functions for fetching template data."""

import json

from sklearn.utils import Bunch

from .datasets_utils import (
    SURFACE,
    _get_reference_info,
    _check_freesurfer_subjid,
    fetch_file,
)


def fetch_fsaverage(
    version="fsaverage", use_local=False, force=False, data_dir=None, verbose=1
):
    """
    Download files for fsaverage FreeSurfer template.

    This dataset contains

    If you used this data, please cite 1_, 2_, 3_.

    Parameters
    ----------
    version : str, optional
        One of {'fsaverage', 'fsaverage3', 'fsaverage4', 'fsaverage5',
        'fsaverage6'}. Default: 'fsaverage'
    use_local : bool, optional
        If True, will attempt to use local FreeSurfer data. Default: False

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with keys ['surf'] where corresponding values
        are length-2 lists downloaded template files (each list composed of
        files for the left and right hemisphere).

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Anders M Dale, Bruce Fischl, and Martin I Sereno. Cortical
        surface-based analysis: i. segmentation and surface reconstruction.
        Neuroimage, 9(2):179\u2013194, 1999.
    .. [2] Bruce Fischl, Martin I Sereno, and Anders M Dale. Cortical
        surface-based analysis: ii: inflation, flattening, and a surface-based
        coordinate system. Neuroimage, 9(2):195\u2013207, 1999.
    .. [3] Bruce Fischl, Martin I Sereno, Roger BH Tootell, and Anders M Dale.
        High-resolution intersubject averaging and a coordinate system for the
        cortical surface. Human brain mapping, 8(4):272\u2013284, 1999.
    """
    versions = ["fsaverage", "fsaverage3", "fsaverage4", "fsaverage5", "fsaverage6"]
    if version not in versions:
        raise ValueError(
            f"The version of fsaverage requested {version} does not "
            f"exist. Must be one of {versions}"
        )

    dataset_name = "tpl-fsaverage"
    _get_reference_info(dataset_name, verbose=verbose)

    keys = ["orig", "white", "smoothwm", "pial", "inflated", "sphere"]

    if use_local:
        try:
            data_dir = _check_freesurfer_subjid(version)[1]
            data = {
                k: SURFACE(
                    data_dir / f"{version}/surf/lh.{k}",
                    data_dir / f"{version}/surf/rh.{k}",
                )
                for k in keys
            }
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Local FreeSurfer data for {version} not found. "
                "Please ensure FreeSurfer is installed and properly set up."
            ) from None
    else:
        fetched = fetch_file(
            dataset_name, keys=version, force=force, data_dir=data_dir, verbose=verbose
        )

        data = {
            k: SURFACE(
                fetched / f"surf/lh.{k}",
                fetched / f"surf/rh.{k}",
            )
            for k in keys
        }

    return Bunch(**data)


def fetch_fsaverage_curated(version="fsaverage", force=False, data_dir=None, verbose=1):
    """
    Download files for fsaverage FreeSurfer template.

    Curated by neuromaps.

    This dataset contains

    If you used this data, please cite 1_, 2_, 3_, 4_.

    Parameters
    ----------
    version : str, optional
        One of {'fsaverage', 'fsaverage4', 'fsaverage5',
        'fsaverage6'}. Default: 'fsaverage'

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with template files.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Anders M Dale, Bruce Fischl, and Martin I Sereno. Cortical
        surface-based analysis: i. segmentation and surface reconstruction.
        Neuroimage, 9(2):179\u2013194, 1999.
    .. [2] Bruce Fischl, Martin I Sereno, and Anders M Dale. Cortical
        surface-based analysis: ii: inflation, flattening, and a surface-based
        coordinate system. Neuroimage, 9(2):195\u2013207, 1999.
    .. [3] Bruce Fischl, Martin I Sereno, Roger BH Tootell, and Anders M Dale.
        High-resolution intersubject averaging and a coordinate system for the
        cortical surface. Human brain mapping, 8(4):272\u2013284, 1999.
    .. [4] Ross D Markello, Justine Y Hansen, Zhen-Qi Liu, Vincent Bazinet,
        Golia Shafiei, Laura E Su\u00e1rez, Nadia Blostein, Jakob Seidlitz,
        Sylvain Baillet, Theodore D Satterthwaite, and others. Neuromaps:
        structural and functional interpretation of brain maps. Nature Methods,
        19(11):1472\u20131479, 2022.
    """
    versions = ["fsaverage", "fsaverage6", "fsaverage5", "fsaverage4"]
    if version not in versions:
        raise ValueError(
            f"The version of fsaverage requested {version} does not "
            f"exist. Must be one of {versions}"
        )

    dataset_name = "tpl-fsaverage_curated"
    _get_reference_info("tpl-fsaverage_curated", verbose=verbose)

    keys = ["white", "pial", "inflated", "sphere", "medial", "sulc", "vaavg"]
    keys_suffix = {
        "white": "white.surf",
        "pial": "pial.surf",
        "inflated": "inflated.surf",
        "sphere": "sphere.surf",
        "medial": "desc-nomedialwall_dparc.label",
        "sulc": "desc-sulc_midthickness.shape",
        "vaavg": "desc-vaavg_midthickness.shape",
    }
    version_density = {
        "fsaverage": "164k",
        "fsaverage6": "41k",
        "fsaverage5": "10k",
        "fsaverage4": "3k",
    }
    density = version_density[version]

    fetched = fetch_file(
        dataset_name, keys=version, force=force, data_dir=data_dir, verbose=verbose
    )

    # deal with default neuromaps directory structure in the archive
    if not fetched.exists():
        import shutil

        shutil.move(fetched.parent / "atlases/fsaverage", fetched)
        shutil.rmtree(fetched.parent / "atlases")

    data = {
        k: SURFACE(
            fetched / f"tpl-fsaverage_den-{density}_hemi-L_{keys_suffix[k]}.gii",
            fetched / f"tpl-fsaverage_den-{density}_hemi-R_{keys_suffix[k]}.gii",
        )
        for k in keys
    }

    return Bunch(**data)


def fetch_hcp_standards(force=False, data_dir=None, verbose=1):
    """
    Fetch HCP standard mesh atlases for converting between FreeSurfer and HCP.

    This dataset contains

    The original file was from 3_, but is no longer available. The archived
    file is available from 4_.

    If you used this data, please cite 1_, 2_.

    Returns
    -------
    standards : str
        Filepath to standard_mesh_atlases directory

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] David C Van Essen, Kamil Ugurbil, Edward Auerbach, Deanna
        Barch,Timothy EJ Behrens, Richard Bucholz, Acer Chang, Liyong Chen,
        Maurizio Corbetta, Sandra W Curtiss, and others. The human connectome
        project: a data acquisition perspective. Neuroimage,
        62(4):2222\u20132231, 2012.
    .. [2] Matthew F Glasser, Stamatios N Sotiropoulos, J Anthony Wilson,
        Timothy S Coalson, Bruce Fischl, Jesper L Andersson, Junqian Xu, Saad
        Jbabdi, Matthew Webster, Jonathan R Polimeni, and others. The minimal
        preprocessing pipelines for the human connectome project. Neuroimage,
        80:105\u2013124, 2013.
    .. [3] http://brainvis.wustl.edu/workbench/standard_mesh_atlases.zip
    .. [4] https://web.archive.org/web/20220121035833/http://brainvis.wustl.edu/workbench/standard_mesh_atlases.zip
    """
    dataset_name = "tpl-hcp_standards"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(
        dataset_name,
        keys="standard_mesh_atlases",
        force=force,
        data_dir=data_dir,
        verbose=verbose,
    )

    return fetched


def fetch_fslr_curated(version="fslr32k", force=False, data_dir=None, verbose=1):
    """
    Download files for HCP fsLR template.

    Curated by neuromaps.

    This dataset contains

    If you used this data, please cite 1_, 2_, 3_.

    Parameters
    ----------
    version : str, optional
        One of {"fslr4k", "fslr8k", "fslr32k", "fslr164k"}. Default: 'fslr32k'

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with template files.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] David C Van Essen, Kamil Ugurbil, Edward Auerbach, Deanna
        Barch,Timothy EJ Behrens, Richard Bucholz, Acer Chang, Liyong Chen,
        Maurizio Corbetta, Sandra W Curtiss, and others. The human connectome
        project: a data acquisition perspective. Neuroimage,
        62(4):2222\u20132231, 2012.
    .. [2] Matthew F Glasser, Stamatios N Sotiropoulos, J Anthony Wilson,
        Timothy S Coalson, Bruce Fischl, Jesper L Andersson, Junqian Xu, Saad
        Jbabdi, Matthew Webster, Jonathan R Polimeni, and others. The minimal
        preprocessing pipelines for the human connectome project. Neuroimage,
        80:105\u2013124, 2013.
    .. [3] Ross D Markello, Justine Y Hansen, Zhen-Qi Liu, Vincent Bazinet,
        Golia Shafiei, Laura E Su\u00e1rez, Nadia Blostein, Jakob Seidlitz,
        Sylvain Baillet, Theodore D Satterthwaite, and others. Neuromaps:
        structural and functional interpretation of brain maps. Nature Methods,
        19(11):1472\u20131479, 2022.
    """
    versions = ["fslr4k", "fslr8k", "fslr32k", "fslr164k"]
    if version not in versions:
        raise ValueError(
            f"The version of fsaverage requested {version} does not "
            f"exist. Must be one of {versions}"
        )

    dataset_name = "tpl-fslr_curated"
    _get_reference_info("tpl-fslr_curated", verbose=verbose)

    keys = [
        "midthickness",
        "inflated",
        "veryinflated",
        "sphere",
        "medial",
        "sulc",
        "vaavg",
    ]
    if version in ["fslr4k", "fslr8k"]:
        keys.remove("veryinflated")
    keys_suffix = {
        "midthickness": "midthickness.surf",
        "inflated": "inflated.surf",
        "veryinflated": "veryinflated.surf",
        "sphere": "sphere.surf",
        "medial": "desc-nomedialwall_dparc.label",
        "sulc": "desc-sulc_midthickness.shape",
        "vaavg": "desc-vaavg_midthickness.shape",
    }
    version_density = {
        "fslr4k": "4k",
        "fslr8k": "8k",
        "fslr32k": "32k",
        "fslr164k": "164k",
    }
    density = version_density[version]

    fetched = fetch_file(
        dataset_name, keys=version, force=force, data_dir=data_dir, verbose=verbose
    )

    # deal with default neuromaps directory structure in the archive
    if not fetched.exists():
        import shutil

        shutil.move(fetched.parent / "atlases/fsLR", fetched)
        shutil.rmtree(fetched.parent / "atlases")

    data = {
        k: SURFACE(
            fetched / f"tpl-fsLR_den-{density}_hemi-L_{keys_suffix[k]}.gii",
            fetched / f"tpl-fsLR_den-{density}_hemi-R_{keys_suffix[k]}.gii",
        )
        for k in keys
    }

    return Bunch(**data)


def fetch_civet(density="41k", version="v1", force=False, data_dir=None, verbose=1):
    """
    Fetch CIVET surface files.

    This dataset contains

    If you used this data, please cite 1_, 2_, 3_.

    Parameters
    ----------
    density : {'41k', '164k'}, optional
        Which density of the CIVET-space geometry files to fetch. The
        high-resolution '164k' surface only exists for version 'v2'
    version : {'v1, 'v2'}, optional
        Which version of the CIVET surfaces to use. Default: 'v2'

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with keys ['mid', 'white'] containing geometry
        files for CIVET surface. Note for version 'v1' the 'mid' and 'white'
        files are identical.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Notes
    -----
    License: https://github.com/aces/CIVET_Full_Project/blob/master/LICENSE

    References
    ----------
    .. [1] Oliver Lyttelton, Maxime Boucher, Steven Robbins, and Alan Evans. An
        unbiased iterative group registration template for cortical surface
        analysis. Neuroimage, 34(4):1535\u20131544, 2007.
    .. [2] Vladimir S Fonov, Alan C Evans, Robert C McKinstry, C Robert Almli,
        and DL Collins. Unbiased nonlinear average age-appropriate brain
        templates from birth to adulthood. NeuroImage, 47:S102, 2009.
    .. [3] Y Ad-Dab'bagh, O Lyttelton, J Muehlboeck, C Lepage, D Einarson, K
        Mok, O Ivanov, R Vincent, J Lerch, and E Fombonne. The civet
        image-processing environment: a fully automated comprehensive pipeline
        for anatomical neuroimaging research. proceedings of the 12th annual
        meeting of the organization for human brain mapping. Florence, Italy,
        pages 2266, 2006.
    """
    densities = ["41k", "164k"]
    if density not in densities:
        raise ValueError(
            f'The density of CIVET requested "{density}" does not exist. '
            f"Must be one of {densities}"
        )
    versions = ["v1", "v2"]
    if version not in versions:
        raise ValueError(
            f'The version of CIVET requested "{version}" does not exist. '
            f"Must be one of {versions}"
        )

    if version == "v1" and density == "164k":
        raise ValueError(
            'The "164k" density CIVET surface only exists for ' 'version "v2"'
        )

    dataset_name = "tpl-civet"
    _get_reference_info(dataset_name, verbose=verbose)

    keys = ["mid", "white"]

    fetched = fetch_file(
        dataset_name,
        keys=[version, "civet" + density],
        force=force,
        data_dir=data_dir,
        verbose=verbose,
    )

    data = {
        k: SURFACE(
            fetched / f"tpl-civet_space-ICBM152_hemi-L_den-{density}_{k}.obj",
            fetched / f"tpl-civet_space-ICBM152_hemi-R_den-{density}_{k}.obj",
        )
        for k in keys
    }
    return Bunch(**data)


def fetch_civet_curated(version="civet41k", force=False, data_dir=None, verbose=1):
    """
    Download files for CIVET template.

    Curated by neuromaps.

    This dataset contains

    If you used this data, please cite 1_, 2_, 3_, 4_.

    Parameters
    ----------
    version : {'civet41k', 'civet164k'}, optional
        Which density of the CIVET-space geometry files to fetch.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with template files.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Notes
    -----
    License: https://github.com/aces/CIVET_Full_Project/blob/master/LICENSE

    References
    ----------
    .. [1] Oliver Lyttelton, Maxime Boucher, Steven Robbins, and Alan Evans. An
        unbiased iterative group registration template for cortical surface
        analysis. Neuroimage, 34(4):1535\u20131544, 2007.
    .. [2] Vladimir S Fonov, Alan C Evans, Robert C McKinstry, C Robert Almli,
        and DL Collins. Unbiased nonlinear average age-appropriate brain
        templates from birth to adulthood. NeuroImage, 47:S102, 2009.
    .. [3] Y Ad-Dab'bagh, O Lyttelton, J Muehlboeck, C Lepage, D Einarson, K
        Mok, O Ivanov, R Vincent, J Lerch, and E Fombonne. The civet
        image-processing environment: a fully automated comprehensive pipeline
        for anatomical neuroimaging research. proceedings of the 12th annual
        meeting of the organization for human brain mapping. Florence, Italy,
        pages 2266, 2006.
    .. [4] Ross D Markello, Justine Y Hansen, Zhen-Qi Liu, Vincent Bazinet,
        Golia Shafiei, Laura E Su\u00e1rez, Nadia Blostein, Jakob Seidlitz,
        Sylvain Baillet, Theodore D Satterthwaite, and others. Neuromaps:
        structural and functional interpretation of brain maps. Nature Methods,
        19(11):1472\u20131479, 2022.
    """
    versions = ["civet41k", "civet164k"]
    if version not in versions:
        raise ValueError(
            f"The version of fsaverage requested {version} does not "
            f"exist. Must be one of {versions}"
        )

    dataset_name = "tpl-civet_curated"
    _get_reference_info("tpl-civet_curated", verbose=verbose)

    keys = [
        "white",
        "midthickness",
        "inflated",
        "veryinflated",
        "sphere",
        "medial",
        "sulc",
        "vaavg",
    ]
    keys_suffix = {
        "white": "white.surf",
        "midthickness": "midthickness.surf",
        "inflated": "inflated.surf",
        "veryinflated": "veryinflated.surf",
        "sphere": "sphere.surf",
        "medial": "desc-nomedialwall_dparc.label",
        "sulc": "desc-sulc_midthickness.shape",
        "vaavg": "desc-vaavg_midthickness.shape",
    }
    version_density = {
        "civet41k": "41k",
        "civet164k": "164k",
    }
    density = version_density[version]

    fetched = fetch_file(
        dataset_name,
        keys=["v2", version],
        force=force,
        data_dir=data_dir,
        verbose=verbose,
    )

    # deal with default neuromaps directory structure in the archive
    if not fetched.exists():
        import shutil

        shutil.move(fetched.parent / "atlases/civet", fetched)
        shutil.rmtree(fetched.parent / "atlases")

    data = {
        k: SURFACE(
            fetched / f"tpl-civet_den-{density}_hemi-L_{keys_suffix[k]}.gii",
            fetched / f"tpl-civet_den-{density}_hemi-R_{keys_suffix[k]}.gii",
        )
        for k in keys
    }

    return Bunch(**data)


def fetch_conte69(force=False, data_dir=None, verbose=1):
    """
    Download files for Van Essen et al., 2012 Conte69 template.

    This dataset contains

    If you used this data, please cite 1_, 2_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with keys ['midthickness', 'inflated',
        'vinflated'], where corresponding values are lists of filepaths to
        downloaded template files.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] David C Van Essen, Kamil Ugurbil, Edward Auerbach, Deanna Barch,
        Timothy EJ Behrens, Richard Bucholz, Acer Chang, Liyong Chen, Maurizio
        Corbetta, Sandra W Curtiss, and others. The human connectome project: a
        data acquisition perspective. Neuroimage, 62(4):2222\u20132231, 2012.
    .. [2] David C Van Essen, Matthew F Glasser, Donna L Dierker, John Harwell,
        and Timothy Coalson. Parcellations and hemispheric asymmetries of human
        cerebral cortex analyzed on surface-based atlases. Cerebral cortex,
        22(10):2241\u20132262, 2012.
    .. [3] http://brainvis.wustl.edu/wiki/index.php//Caret:Atlases/Conte69_Atlas
    """
    dataset_name = "tpl-conte69"
    _get_reference_info(dataset_name, verbose=verbose)

    keys = ["midthickness", "inflated", "vinflated"]

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    data = {
        k: SURFACE(
            fetched / f"tpl-conte69_space-MNI305_variant-fsLR32k_{k}.L.surf.gii",
            fetched / f"tpl-conte69_space-MNI305_variant-fsLR32k_{k}.R.surf.gii",
        )
        for k in keys
    }
    data["info"] = json.load(open(fetched / "template_description.json", "r"))

    return Bunch(**data)


def fetch_yerkes19(force=False, data_dir=None, verbose=1):
    """
    Download files for Donahue et al., 2016 Yerkes19 template.

    This dataset contains

    If you used this data, please cite 1_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with keys ['midthickness', 'inflated',
        'vinflated'], where corresponding values are lists of filepaths to
        downloaded template files.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Chad J Donahue, Stamatios N Sotiropoulos, Saad Jbabdi, Moises
        Hernandez-Fernandez, Timothy E Behrens, Tim B Dyrby, Timothy Coalson,
        Henry Kennedy, Kenneth Knoblauch, David C Van Essen, and others. Using
        diffusion tractography to predict cortical connection strength and
        distance: a quantitative comparison with tracers in the monkey. Journal
        of Neuroscience, 36(25):6758\u20136770, 2016.
    .. [2] https://balsa.wustl.edu/reference/show/976nz
    """
    dataset_name = "tpl-yerkes19"
    _get_reference_info(dataset_name, verbose=verbose)

    keys = ["midthickness", "inflated", "vinflated"]

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    data = {
        k: SURFACE(
            fetched / f"tpl-yerkes19_space-fsLR32k_{k}.L.surf.gii",
            fetched / f"tpl-yerkes19_space-fsLR32k_{k}.R.surf.gii",
        )
        for k in keys
    }

    return Bunch(**data)



"""Functions for pyvista-based plotting."""

from pathlib import Path

import nibabel as nib
import numpy as np

try:
    import pyvista as pv
except ImportError:
    _has_pyvista = False
else:
    _has_pyvista = True

from netneurotools.datasets import (
    fetch_civet_curated,
    fetch_fsaverage_curated,
    fetch_fslr_curated,
)


def _pv_fetch_template(template, surf="inflated", data_dir=None, verbose=0):
    if template in ["fsaverage", "fsaverage6", "fsaverage5", "fsaverage4"]:
        _fetch_curr_tpl = fetch_fsaverage_curated
    elif template in ["fslr4k", "fslr8k", "fslr32k", "fslr164k"]:
        _fetch_curr_tpl = fetch_fslr_curated
    elif template in ["civet41k", "civet164k"]:
        _fetch_curr_tpl = fetch_civet_curated
    else:
        raise ValueError(f"Unknown template: {template}")

    curr_tpl_surf = _fetch_curr_tpl(
        version=template, data_dir=data_dir, verbose=verbose
    )[surf]

    return curr_tpl_surf


def _pv_make_surface(template, surf="inflated", hemi=None, data_dir=None, verbose=0):
    curr_tpl_surf = _pv_fetch_template(
        template=template, surf=surf, data_dir=data_dir, verbose=verbose
    )

    def _gifti_to_polydata(gifti_file):
        vertices, faces = nib.load(gifti_file).agg_data()
        return pv.PolyData(
            vertices, np.c_[np.ones((faces.shape[0],), dtype=int) * 3, faces]
        )

    if hemi == "L":
        return _gifti_to_polydata(curr_tpl_surf.L)
    elif hemi == "R":
        return _gifti_to_polydata(curr_tpl_surf.R)
    else:
        return (
            _gifti_to_polydata(curr_tpl_surf.L),
            _gifti_to_polydata(curr_tpl_surf.R),
        )


def _mask_medial_wall(data, template, hemi=None, data_dir=None, verbose=0):
    curr_medial = _pv_fetch_template(
        template=template, surf="medial", data_dir=data_dir, verbose=verbose
    )
    if isinstance(data, tuple):
        curr_medial_data = (
            nib.load(curr_medial.L).agg_data(),
            nib.load(curr_medial.R).agg_data(),
        )
        # convert to float for NaNs
        # caution: might cause issues
        ret_L = data[0].astype(float)
        ret_R = data[1].astype(float)
        ret_L[np.where(1 - curr_medial_data[0])] = np.nan
        ret_R[np.where(1 - curr_medial_data[1])] = np.nan
        ret = (ret_L, ret_R)
    else:
        if hemi == "L":
            curr_medial_data = nib.load(curr_medial.L).agg_data()
        elif hemi == "R":
            curr_medial_data = nib.load(curr_medial.R).agg_data()
        else:
            curr_medial_data = np.concatenate(
                [
                    nib.load(curr_medial.L).agg_data(),
                    nib.load(curr_medial.R).agg_data(),
                ],
                axis=1,
            )
        ret = data.copy()
        ret[np.where(1 - curr_medial_data)] = np.nan
    return ret


def pv_plot_surface(
    vertex_data,
    template,
    surf="inflated",
    hemi="both",
    layout="default",
    mask_medial=True,
    cmap="viridis",
    clim=None,
    zoom_ratio=1.25,
    show_colorbar=True,
    cbar_title=None,
    show_plot=True,
    jupyter_backend="html",
    lighting_style="default",
    save_fig=None,
    plotter_kws=None,
    mesh_kws=None,
    cbar_kws=None,
    silhouette_kws=None,
    data_dir=None,
    verbose=0,
):
    """
    Plot surface data using PyVista.

    Parameters
    ----------
    vertex_data : array-like or tuple of array-like
        Data array(s) to be plotted on the surface. If `hemi` is "both", this
        should be a tuple of two arrays. Otherwise, a single array.
    template : str
        Template to use for plotting. Options include 'fsaverage', 'fsaverage6',
        'fsaverage5', 'fsaverage4', 'fslr4k', 'fslr8k', 'fslr32k', 'fslr164k',
        'civet41k', 'civet164k'.
    surf : str, optional
        Surface to plot. Default is 'inflated'.
    hemi : str, optional
        Hemisphere to plot. Options include 'L', 'R', 'both'. Default is 'both'.
    layout : str, optional
        Layout of the plot. Options include 'default', 'single', 'row', 'column'.
        Default is 'default'.
    mask_medial : bool, optional
        Mask medial wall. Default is True.
    cmap : str, optional
        Colormap to use. Default is 'viridis'.
    clim : tuple, optional
        Colorbar limits. If None, will be set to 2.5th and 97.5th percentiles.
        Default is None.
    zoom_ratio : float, optional
        Zoom ratio for the camera. Default is 1.25.
    show_colorbar : bool, optional
        Whether to show the colorbar. Default is True.
    cbar_title : str, optional
        Title for the colorbar. Default is None.
    show_plot : bool, optional
        Whether to show the plot. Default is True.
    jupyter_backend : str, optional
        Jupyter backend to use. See `PyVista documentation
        <https://docs.pyvista.org/user-guide/jupyter/index.html#pyvista.set_jupyter_backend>`_
        for more details. Default is 'html'.
    lighting_style : str, optional
        Lighting style to use. Options include 'default', 'lightkit', 'threelights',
        'silhouette', 'metallic', 'plastic', 'shiny', 'glossy', 'ambient', 'plain'.
        Default is 'default'.
    save_fig : str or Path, optional
        Path (include file name) to save the figure. Default is None.

    Returns
    -------
    pl : :class:`pyvista.Plotter`
        PyVista plotter object.

    Other Parameters
    ----------------
    plotter_kws : dict, optional
        Additional keyword arguments to pass to the `PyVista plotter
        <https://docs.pyvista.org/api/plotting/_autosummary/pyvista.plotter>`_.
        Default is None.
    mesh_kws : dict, optional
        Additional keyword arguments to pass to the `PyVista mesh
        <https://docs.pyvista.org/api/plotting/_autosummary/pyvista.plotter.add_mesh>`_.
        Default is None.
    cbar_kws : dict, optional
        Additional keyword arguments to pass to the `PyVista colorbar
        <https://docs.pyvista.org/api/plotting/_autosummary/pyvista.plotter.add_scalar_bar>`_.
        Default is None.
    silhouette_kws : dict, optional
        Additional keyword arguments to pass to the `PyVista silhouette
        <https://docs.pyvista.org/api/plotting/_autosummary/pyvista.plotter.add_silhouette>`_.
        Default is None.
    data_dir : str or Path, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 0
    """
    if not _has_pyvista:
        raise ImportError("PyVista is required for this function")

    # setup data
    # could be a single array or a tuple of two arrays
    if hemi == "both":  # both hemispheres
        surf_pair = _pv_make_surface(
            template=template, surf=surf, data_dir=data_dir, verbose=verbose
        )
        if len(vertex_data) == 2:  # tuple or list of two arrays
            # check if data length matches number of vertices
            if not all(len(vertex_data[i]) == surf_pair[i].n_points for i in range(2)):
                raise ValueError("Data length mismatch")
        else:  # combined array
            # check if data length matches number of vertices
            if len(vertex_data) != surf_pair[0].n_points + surf_pair[1].n_points:
                raise ValueError("Data length mismatch")
            # convert long array to tuple
            vertex_data = (
                vertex_data[: surf_pair[0].n_points],
                vertex_data[surf_pair[0].n_points :],
            )

        if mask_medial:
            vertex_data = _mask_medial_wall(
                vertex_data, template, hemi=None, data_dir=data_dir, verbose=verbose
            )
        surf_pair[0].point_data["vertex_data"] = vertex_data[0]
        surf_pair[1].point_data["vertex_data"] = vertex_data[1]
    elif hemi in ["L", "R"]:
        # single hemisphere
        surf = _pv_make_surface(
            template=template, surf=surf, hemi=hemi, data_dir=data_dir, verbose=verbose
        )
        if len(vertex_data) != surf.n_points:
            raise ValueError("Data length mismatch")

        if mask_medial:
            vertex_data = _mask_medial_wall(
                vertex_data, template, hemi=hemi, data_dir=data_dir, verbose=verbose
            )
        surf.point_data["vertex_data"] = vertex_data
    else:
        raise ValueError(f"Unknown hemi: {hemi}")

    # setup plotter shape based on layout
    if layout == "default":
        if hemi == "both":
            plotter_shape = (2, 2)
        else:
            plotter_shape = (1, 2)
    elif layout == "single":
        plotter_shape = (1, 1)
    elif layout == "row":
        if hemi == "both":
            plotter_shape = (1, 4)
        else:
            plotter_shape = (2, 1)
    elif layout == "column":
        if hemi == "both":
            plotter_shape = (4, 1)
        else:
            plotter_shape = (1, 2)
    else:
        raise ValueError(f"Unknown layout: {layout}")

    # setup color limits
    if clim is not None:
        _vmin, _vmax = clim
    else:
        if len(vertex_data) == 2:
            _values = np.c_[vertex_data[0], vertex_data[1]]
        else:
            _values = vertex_data
        _vmin, _vmax = np.nanpercentile(_values, [2.5, 97.5])

    # default plotter settings
    plotter_settings = dict(
        window_size=(700 * plotter_shape[1], 500 * plotter_shape[0]),
        border=False,
        lighting="three lights",
    )
    # notebook plotting
    if jupyter_backend is not None:
        plotter_settings.update(dict(notebook=True, off_screen=True))

    # default mesh settings
    mesh_settings = dict(
        scalars="vertex_data",
        smooth_shading=True,
        cmap=cmap,
        clim=(_vmin, _vmax),
        show_scalar_bar=False,
    )

    # lighting styles
    lighting_style_keys = ["ambient", "diffuse", "specular", "specular_power"]
    lighting_style_presets = {
        "metallic": [0.1, 0.3, 1.0, 10],
        "plastic": [0.3, 0.4, 0.3, 5],
        "shiny": [0.2, 0.6, 0.8, 50],
        "glossy": [0.1, 0.7, 0.9, 90],
        "ambient": [0.8, 0.1, 0.0, 1],
        "plain": [0.1, 1.0, 0.05, 5],
    }

    if lighting_style in ["default", "lightkit"]:
        mesh_settings["lighting"] = "light kit"
    elif lighting_style == "threelights":
        mesh_settings["lighting"] = "three lights"
    elif lighting_style == "silhouette":
        mesh_settings["lighting"] = "light kit"
    elif lighting_style in lighting_style_presets.keys():
        mesh_settings.update(
            {
                k: v
                for k, v in zip(
                    lighting_style_keys, lighting_style_presets[lighting_style]
                )
            }
        )
        mesh_settings["lighting"] = "light kit"
    else:
        raise ValueError(f"Unknown lighting style: {lighting_style}")

    # default colorbar settings
    cbar_settings = dict(
        title=cbar_title,
        n_labels=2,
        label_font_size=20,
        title_font_size=24,
        font_family="arial",
        height=0.15,
    )

    # default silhouette settings
    silhouette_settings = dict(color="white", feature_angle=40)

    # update if provided with custom settings
    if plotter_kws is not None:
        plotter_settings.update(plotter_kws)
    if mesh_kws is not None:
        mesh_settings.update(mesh_kws)
    if cbar_kws is not None:
        cbar_settings.update(cbar_kws)
    if silhouette_kws is not None:
        silhouette_settings.update(silhouette_kws)

    pl = pv.Plotter(shape=plotter_shape, **plotter_settings)

    if layout == "single":  # single panel (1, 1)
        if hemi == "both":
            _surf = surf_pair[0].rotate_z(180)
        pl.subplot(0, 0)
        pl.add_mesh(_surf, **mesh_settings)
        pl.camera_position = "yz"
        pl.zoom_camera(zoom_ratio)
        if lighting_style == "silhouette":
            pl.add_silhouette(_surf, **silhouette_settings)
    else:  # multiple panels
        if hemi == "both":  # both hemi, 4 panels
            if layout == "default":
                _pos = [(0, 0), (0, 1), (1, 0), (1, 1)]
            elif layout == "row":
                _pos = [(0, 0), (0, 2), (0, 1), (0, 3)]
            elif layout == "column":
                _pos = [(0, 0), (2, 0), (1, 0), (3, 0)]
            else:
                raise ValueError(f"Unknown layout: {layout}")
            _surf_list = [
                surf_pair[0].rotate_z(180),
                surf_pair[1],
                surf_pair[0],
                surf_pair[1].rotate_z(180),
            ]
            for _xy, _surf in zip(_pos, _surf_list):
                pl.subplot(*_xy)
                pl.add_mesh(_surf, **mesh_settings)
                pl.camera_position = "yz"
                pl.zoom_camera(zoom_ratio)
                if lighting_style == "silhouette":
                    pl.add_silhouette(_surf, **silhouette_settings)
        else:  # single hemi, 2 panels
            if layout == "default":
                _pos = [(0, 0), (0, 1)]
            elif layout == "row":
                _pos = [(0, 0), (0, 1)]
            elif layout == "column":
                _pos = [(0, 0), (1, 0)]
            else:
                raise ValueError(f"Unknown layout: {layout}")

            if hemi == "L":
                _surf_list = [surf.rotate_z(180), surf]
            else:
                _surf_list = [surf, surf.rotate_z(180)]

            for _xy, _surf in zip(_pos, _surf_list):
                pl.subplot(*_xy)
                pl.add_mesh(_surf, **mesh_settings)
                pl.camera_position = "yz"
                pl.zoom_camera(zoom_ratio)
                if lighting_style == "silhouette":
                    pl.add_silhouette(_surf, **silhouette_settings)

    if show_colorbar:
        cbar = pl.add_scalar_bar(**cbar_settings)
        cbar.GetLabelTextProperty().SetItalic(True)

    # setting the headlight (by default applied to all scenes)
    if lighting_style in ["default", "silhouette"] + list(
        lighting_style_presets.keys()
    ):
        light = pv.Light(light_type="headlight", intensity=0.2)
        pl.add_light(light)

    if show_plot:
        if jupyter_backend is not None:
            pl.show(jupyter_backend=jupyter_backend)
        else:
            pl.show()

    if save_fig is not None:
        _fname = Path(save_fig)
        if _fname.suffix in [".png", ".jpeg", ".jpg", ".bmp", ".tif", ".tiff"]:
            pl.screenshot(_fname, return_img=False)
        elif _fname.suffix in [".svg", ".eps", ".ps", ".pdf", ".tex"]:
            pl.save_graphic(_fname)
        else:
            raise ValueError(f"Unknown file format: {save_fig}")

    return pl
